import logging
import tempfile

from sqlalchemy import (
    false,
    or_,
)

from galaxy import util
from galaxy.model.base import transaction
from galaxy.util import inflector
from galaxy.web.form_builder import SelectField
from tool_shed.metadata import metadata_generator
from tool_shed.repository_types import util as rt_util
from tool_shed.repository_types.metadata import TipOnly
from tool_shed.structured_app import ToolShedApp
from tool_shed.util import (
    basic_util,
    common_util,
    hg_util,
    metadata_util,
    repository_util,
    shed_util_common as suc,
    tool_util,
)

log = logging.getLogger(__name__)


class RepositoryMetadataManager(metadata_generator.MetadataGenerator):
    def __init__(
        self,
        app: ToolShedApp,
        user,
        repository=None,
        changeset_revision=None,
        repository_clone_url=None,
        shed_config_dict=None,
        relative_install_dir=None,
        repository_files_dir=None,
        resetting_all_metadata_on_repository=False,
        updating_installed_repository=False,
        persist=False,
        metadata_dict=None,
    ):
        super().__init__(
            app,
            repository,
            changeset_revision,
            repository_clone_url,
            shed_config_dict,
            relative_install_dir,
            repository_files_dir,
            resetting_all_metadata_on_repository,
            updating_installed_repository,
            persist,
            metadata_dict=metadata_dict,
            user=user,
        )
        self.app = app
        self.user = user
        # Repository metadata comparisons for changeset revisions.
        self.EQUAL = "equal"
        self.NO_METADATA = "no metadata"
        self.NOT_EQUAL_AND_NOT_SUBSET = "not equal and not subset"
        self.SUBSET = "subset"
        self.SUBSET_VALUES = [self.EQUAL, self.SUBSET]

    def add_tool_versions(self, id, repository_metadata, changeset_revisions):
        # Build a dictionary of { 'tool id' : 'parent tool id' } pairs for each tool in repository_metadata.
        metadata = repository_metadata.metadata
        tool_versions_dict = {}
        for tool_dict in metadata.get("tools", []):
            # We have at least 2 changeset revisions to compare tool guids and tool ids.
            parent_id = self.get_parent_id(
                id, tool_dict["id"], tool_dict["version"], tool_dict["guid"], changeset_revisions
            )
            tool_versions_dict[tool_dict["guid"]] = parent_id
        if tool_versions_dict:
            repository_metadata.tool_versions = tool_versions_dict
            self.sa_session.add(repository_metadata)
            session = self.sa_session()
            with transaction(session):
                session.commit()

    def build_repository_ids_select_field(
        self, name="repository_ids", multiple=True, display="checkboxes", my_writable=False
    ):
        """Generate the current list of repositories for resetting metadata."""
        repositories_select_field = SelectField(name=name, multiple=multiple, display=display)
        query = self.get_query_for_setting_metadata_on_repositories(my_writable=my_writable, order=True)
        for repository in query:
            owner = str(repository.user.username)
            option_label = f"{str(repository.name)} ({owner})"
            option_value = f"{self.app.security.encode_id(repository.id)}"
            repositories_select_field.add_option(option_label, option_value)
        return repositories_select_field

    def clean_repository_metadata(self, changeset_revisions):
        # Delete all repository_metadata records associated with the repository that have
        # a changeset_revision that is not in changeset_revisions.  We sometimes see multiple
        # records with the same changeset revision value - no idea how this happens. We'll
        # assume we can delete the older records, so we'll order by update_time descending and
        # delete records that have the same changeset_revision we come across later.
        changeset_revisions_checked = []
        for repository_metadata in (
            self.sa_session.query(self.app.model.RepositoryMetadata)
            .filter(self.app.model.RepositoryMetadata.table.c.repository_id == self.repository.id)
            .order_by(
                self.app.model.RepositoryMetadata.table.c.changeset_revision,
                self.app.model.RepositoryMetadata.table.c.update_time.desc(),
            )
        ):
            changeset_revision = repository_metadata.changeset_revision
            if changeset_revision in changeset_revisions_checked or changeset_revision not in changeset_revisions:
                self.sa_session.delete(repository_metadata)
                session = self.sa_session()
                with transaction(session):
                    session.commit()

    def compare_changeset_revisions(self, ancestor_changeset_revision, ancestor_metadata_dict):
        """
        Compare the contents of two changeset revisions to determine if a new repository
        metadata revision should be created.
        """
        # The metadata associated with ancestor_changeset_revision is ancestor_metadata_dict.
        # This changeset_revision is an ancestor of self.changeset_revision which is associated
        # with self.metadata_dict.  A new repository_metadata record will be created only
        # when this method returns the constant value self.NOT_EQUAL_AND_NOT_SUBSET.
        ancestor_datatypes = ancestor_metadata_dict.get("datatypes", [])
        ancestor_tools = ancestor_metadata_dict.get("tools", [])
        ancestor_guids = [tool_dict["guid"] for tool_dict in ancestor_tools]
        ancestor_guids.sort()
        ancestor_readme_files = ancestor_metadata_dict.get("readme_files", [])
        ancestor_repository_dependencies_dict = ancestor_metadata_dict.get("repository_dependencies", {})
        ancestor_repository_dependencies = ancestor_repository_dependencies_dict.get("repository_dependencies", [])
        ancestor_tool_dependencies = ancestor_metadata_dict.get("tool_dependencies", {})
        ancestor_workflows = ancestor_metadata_dict.get("workflows", [])
        ancestor_data_manager = ancestor_metadata_dict.get("data_manager", {})
        current_datatypes = self.metadata_dict.get("datatypes", [])
        current_tools = self.metadata_dict.get("tools", [])
        current_guids = [tool_dict["guid"] for tool_dict in current_tools]
        current_guids.sort()
        current_readme_files = self.metadata_dict.get("readme_files", [])
        current_repository_dependencies_dict = self.metadata_dict.get("repository_dependencies", {})
        current_repository_dependencies = current_repository_dependencies_dict.get("repository_dependencies", [])
        current_tool_dependencies = self.metadata_dict.get("tool_dependencies", {})
        current_workflows = self.metadata_dict.get("workflows", [])
        current_data_manager = self.metadata_dict.get("data_manager", {})
        # Handle case where no metadata exists for either changeset.
        no_datatypes = not ancestor_datatypes and not current_datatypes
        no_readme_files = not ancestor_readme_files and not current_readme_files
        no_repository_dependencies = not ancestor_repository_dependencies and not current_repository_dependencies
        no_tool_dependencies = not ancestor_tool_dependencies and not current_tool_dependencies
        no_tools = not ancestor_guids and not current_guids
        no_workflows = not ancestor_workflows and not current_workflows
        no_data_manager = not ancestor_data_manager and not current_data_manager
        if (
            no_datatypes
            and no_readme_files
            and no_repository_dependencies
            and no_tool_dependencies
            and no_tools
            and no_workflows
            and no_data_manager
        ):
            return self.NO_METADATA
        # Uncomment the following if we decide that README files should affect how installable
        # repository revisions are defined.  See the NOTE in self.compare_readme_files().
        # readme_file_comparision = self.compare_readme_files( ancestor_readme_files, current_readme_files )
        repository_dependency_comparison = self.compare_repository_dependencies(
            ancestor_repository_dependencies, current_repository_dependencies
        )
        tool_dependency_comparison = self.compare_tool_dependencies(
            ancestor_tool_dependencies, current_tool_dependencies
        )
        workflow_comparison = self.compare_workflows(ancestor_workflows, current_workflows)
        datatype_comparison = self.compare_datatypes(ancestor_datatypes, current_datatypes)
        data_manager_comparison = self.compare_data_manager(ancestor_data_manager, current_data_manager)
        # Handle case where all metadata is the same.
        if (
            ancestor_guids == current_guids
            and repository_dependency_comparison == self.EQUAL
            and tool_dependency_comparison == self.EQUAL
            and workflow_comparison == self.EQUAL
            and datatype_comparison == self.EQUAL
            and data_manager_comparison == self.EQUAL
        ):
            return self.EQUAL
        # Handle case where ancestor metadata is a subset of current metadata.
        # readme_file_is_subset = readme_file_comparision in [ self.EQUAL, self.SUBSET ]
        repository_dependency_is_subset = repository_dependency_comparison in self.SUBSET_VALUES
        tool_dependency_is_subset = tool_dependency_comparison in self.SUBSET_VALUES
        workflow_dependency_is_subset = workflow_comparison in self.SUBSET_VALUES
        datatype_is_subset = datatype_comparison in self.SUBSET_VALUES
        datamanager_is_subset = data_manager_comparison in self.SUBSET_VALUES
        if (
            repository_dependency_is_subset
            and tool_dependency_is_subset
            and workflow_dependency_is_subset
            and datatype_is_subset
            and datamanager_is_subset
        ):
            is_subset = True
            for guid in ancestor_guids:
                if guid not in current_guids:
                    is_subset = False
                    break
            if is_subset:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_data_manager(self, ancestor_metadata, current_metadata):
        """Determine if ancestor_metadata is the same as or a subset of current_metadata for data_managers."""

        def __data_manager_dict_to_tuple_list(metadata_dict):
            # we do not check tool_guid or tool conf file name
            return set(
                sorted(
                    (
                        name,
                        tuple(sorted(value.get("data_tables", []))),
                        value.get("guid"),
                        value.get("version"),
                        value.get("name"),
                        value.get("id"),
                    )
                    for name, value in metadata_dict.items()
                )
            )

        # only compare valid entries, any invalid entries are ignored
        ancestor_metadata = __data_manager_dict_to_tuple_list(ancestor_metadata.get("data_managers", {}))
        current_metadata = __data_manager_dict_to_tuple_list(current_metadata.get("data_managers", {}))
        # use set comparisons
        if ancestor_metadata.issubset(current_metadata):
            if ancestor_metadata == current_metadata:
                return self.EQUAL
            return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_datatypes(self, ancestor_datatypes, current_datatypes):
        """Determine if ancestor_datatypes is the same as or a subset of current_datatypes."""
        # Each datatype dict looks something like:
        # {"dtype": "galaxy.datatypes.images:Image", "extension": "pdf", "mimetype": "application/pdf"}
        if len(ancestor_datatypes) <= len(current_datatypes):
            for ancestor_datatype in ancestor_datatypes:
                # Currently the only way to differentiate datatypes is by name.
                ancestor_datatype_dtype = ancestor_datatype["dtype"]
                ancestor_datatype_extension = ancestor_datatype["extension"]
                ancestor_datatype_mimetype = ancestor_datatype.get("mimetype", None)
                found_in_current = False
                for current_datatype in current_datatypes:
                    if (
                        current_datatype["dtype"] == ancestor_datatype_dtype
                        and current_datatype["extension"] == ancestor_datatype_extension
                        and current_datatype.get("mimetype", None) == ancestor_datatype_mimetype
                    ):
                        found_in_current = True
                        break
                if not found_in_current:
                    return self.NOT_EQUAL_AND_NOT_SUBSET
            if len(ancestor_datatypes) == len(current_datatypes):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_readme_files(self, ancestor_readme_files, current_readme_files):
        """Determine if ancestor_readme_files is equal to or a subset of current_readme_files."""
        # NOTE: Although repository README files are considered a Galaxy utility similar to tools,
        # repository dependency definition files, etc., we don't define installable repository revisions
        # based on changes to README files.  To understand why, consider the following scenario:
        # 1. Upload the filtering tool to a new repository - this will result in installable revision 0.
        # 2. Upload a README file to the repository - this will move the installable revision from revision
        #    0 to revision 1.
        # 3. Delete the README file from the repository - this will move the installable revision from
        #    revision 1 to revision 2.
        # The above scenario is the current behavior, and that is why this method is not currently called.
        # This method exists only in case we decide to change this current behavior.
        # The lists of readme files looks something like: ["database/community_files/000/repo_2/readme.txt"]
        if len(ancestor_readme_files) <= len(current_readme_files):
            for ancestor_readme_file in ancestor_readme_files:
                if ancestor_readme_file not in current_readme_files:
                    return self.NOT_EQUAL_AND_NOT_SUBSET
            if len(ancestor_readme_files) == len(current_readme_files):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_repository_dependencies(self, ancestor_repository_dependencies, current_repository_dependencies):
        """
        Determine if ancestor_repository_dependencies is the same as or a subset of
        current_repository_dependencies.
        """
        # The list of repository_dependencies looks something like:
        # [["http://localhost:9009", "emboss_datatypes", "test", "ab03a2a5f407", "False", "False"]].
        # Create a string from each tuple in the list for easier comparison.
        if len(ancestor_repository_dependencies) <= len(current_repository_dependencies):
            for ancestor_tup in ancestor_repository_dependencies:
                (
                    a_tool_shed,
                    a_repo_name,
                    a_repo_owner,
                    a_changeset_revision,
                    a_prior_installation_required,
                    a_only_if_compiling_contained_td,
                ) = ancestor_tup
                cleaned_a_tool_shed = common_util.remove_protocol_from_tool_shed_url(a_tool_shed)
                found_in_current = False
                for current_tup in current_repository_dependencies:
                    (
                        c_tool_shed,
                        c_repo_name,
                        c_repo_owner,
                        c_changeset_revision,
                        c_prior_installation_required,
                        c_only_if_compiling_contained_td,
                    ) = current_tup
                    cleaned_c_tool_shed = common_util.remove_protocol_from_tool_shed_url(c_tool_shed)
                    if (
                        cleaned_c_tool_shed == cleaned_a_tool_shed
                        and c_repo_name == a_repo_name
                        and c_repo_owner == a_repo_owner
                        and c_changeset_revision == a_changeset_revision
                        and util.string_as_bool(c_prior_installation_required)
                        == util.string_as_bool(a_prior_installation_required)
                        and util.string_as_bool(c_only_if_compiling_contained_td)
                        == util.string_as_bool(a_only_if_compiling_contained_td)
                    ):
                        found_in_current = True
                        break
                if not found_in_current:
                    # In some cases, the only difference between a dependency definition in the lists
                    # is the changeset_revision value.  We'll check to see if this is the case, and if
                    # the defined dependency is a repository that has metadata set only on its tip.
                    if not self.different_revision_defines_tip_only_repository_dependency(
                        ancestor_tup, current_repository_dependencies
                    ):
                        return self.NOT_EQUAL_AND_NOT_SUBSET
                    return self.SUBSET
            if len(ancestor_repository_dependencies) == len(current_repository_dependencies):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_tool_dependencies(self, ancestor_tool_dependencies, current_tool_dependencies):
        """
        Determine if ancestor_tool_dependencies is the same as or a subset of current_tool_dependencies.
        """
        # The tool_dependencies dictionary looks something like:
        # {'bwa/0.5.9': {'readme': 'some string', 'version': '0.5.9', 'type': 'package', 'name': 'bwa'}}
        if len(ancestor_tool_dependencies) <= len(current_tool_dependencies):
            for ancestor_td_key in ancestor_tool_dependencies.keys():
                if ancestor_td_key in current_tool_dependencies:
                    # The only values that could have changed between the 2 dictionaries are the
                    # "readme" or "type" values.  Changing the readme value makes no difference.
                    # Changing the type will change the installation process, but for now we'll
                    # assume it was a typo, so new metadata shouldn't be generated.
                    continue
                else:
                    return self.NOT_EQUAL_AND_NOT_SUBSET
            # At this point we know that ancestor_tool_dependencies is at least a subset of current_tool_dependencies.
            if len(ancestor_tool_dependencies) == len(current_tool_dependencies):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_workflows(self, ancestor_workflows, current_workflows):
        """
        Determine if ancestor_workflows is the same as current_workflows or if ancestor_workflows
        is a subset of current_workflows.
        """
        if len(ancestor_workflows) <= len(current_workflows):
            for ancestor_workflow_tup in ancestor_workflows:
                # ancestor_workflows is a list of tuples where each contained tuple is
                # [ <relative path to the .ga file in the repository>, <exported workflow dict> ]
                ancestor_workflow_dict = ancestor_workflow_tup[1]
                # Currently the only way to differentiate workflows is by name.
                ancestor_workflow_name = ancestor_workflow_dict["name"]
                num_ancestor_workflow_steps = len(ancestor_workflow_dict["steps"])
                found_in_current = False
                for current_workflow_tup in current_workflows:
                    current_workflow_dict = current_workflow_tup[1]
                    # Assume that if the name and number of steps are euqal, then the workflows
                    # are the same.  Of course, this may not be true...
                    if (
                        current_workflow_dict["name"] == ancestor_workflow_name
                        and len(current_workflow_dict["steps"]) == num_ancestor_workflow_steps
                    ):
                        found_in_current = True
                        break
                if not found_in_current:
                    return self.NOT_EQUAL_AND_NOT_SUBSET
            if len(ancestor_workflows) == len(current_workflows):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def create_or_update_repository_metadata(self, changeset_revision, metadata_dict):
        """Create or update a repository_metadata record in the tool shed."""
        has_repository_dependencies = False
        has_repository_dependencies_only_if_compiling_contained_td = False
        includes_tools = False
        includes_tool_dependencies = False
        if metadata_dict:
            repository_dependencies_dict = metadata_dict.get("repository_dependencies", {})
            repository_dependencies = repository_dependencies_dict.get("repository_dependencies", [])
            (
                has_repository_dependencies,
                has_repository_dependencies_only_if_compiling_contained_td,
            ) = repository_util.get_repository_dependency_types(repository_dependencies)
            if "tools" in metadata_dict:
                includes_tools = True
            if "tool_dependencies" in metadata_dict:
                includes_tool_dependencies = True
        if (
            has_repository_dependencies
            or has_repository_dependencies_only_if_compiling_contained_td
            or includes_tools
            or includes_tool_dependencies
        ):
            downloadable = True
        else:
            downloadable = False
        repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
            self.app, self.app.security.encode_id(self.repository.id), changeset_revision
        )
        if repository_metadata:
            repository_metadata.metadata = metadata_dict
            repository_metadata.downloadable = downloadable
            repository_metadata.has_repository_dependencies = has_repository_dependencies
            repository_metadata.includes_datatypes = False
            repository_metadata.includes_tools = includes_tools
            repository_metadata.includes_tool_dependencies = includes_tool_dependencies
            repository_metadata.includes_workflows = False
        else:
            repository_metadata = self.app.model.RepositoryMetadata(
                repository_id=self.repository.id,
                changeset_revision=changeset_revision,
                metadata=metadata_dict,
                downloadable=downloadable,
                has_repository_dependencies=has_repository_dependencies,
                includes_datatypes=False,
                includes_tools=includes_tools,
                includes_tool_dependencies=includes_tool_dependencies,
                includes_workflows=False,
            )
        assert repository_metadata
        # Always set the default values for the following columns.  When resetting all metadata
        # on a repository this will reset the values.
        repository_metadata.missing_test_components = False
        self.sa_session.add(repository_metadata)
        session = self.sa_session()
        with transaction(session):
            session.commit()

        return repository_metadata

    def different_revision_defines_tip_only_repository_dependency(self, rd_tup, repository_dependencies):
        """
        Determine if the only difference between rd_tup and a dependency definition in the list of
        repository_dependencies is the changeset_revision value.
        """
        (
            rd_tool_shed,
            rd_name,
            rd_owner,
            rd_changeset_revision,
            rd_prior_installation_required,
            rd_only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(rd_tup)
        cleaned_rd_tool_shed = common_util.remove_protocol_from_tool_shed_url(rd_tool_shed)
        for repository_dependency in repository_dependencies:
            (
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ) = common_util.parse_repository_dependency_tuple(repository_dependency)
            cleaned_tool_shed = common_util.remove_protocol_from_tool_shed_url(tool_shed)
            if cleaned_rd_tool_shed == cleaned_tool_shed and rd_name == name and rd_owner == owner:
                # Determine if the repository represented by the dependency tuple is an instance of the repository type TipOnly.
                required_repository = repository_util.get_repository_by_name_and_owner(self.app, name, owner)
                repository_type_class = self.app.repository_types_registry.get_class_by_label(required_repository.type)
                return isinstance(repository_type_class, TipOnly)
        return False

    def get_parent_id(self, id, old_id, version, guid, changeset_revisions):
        parent_id = None
        # Compare from most recent to oldest.
        changeset_revisions.reverse()
        for changeset_revision in changeset_revisions:
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                self.app, id, changeset_revision
            )
            assert repository_metadata
            metadata = repository_metadata.metadata
            tools_dicts = metadata.get("tools", [])
            for tool_dict in tools_dicts:
                if tool_dict["guid"] == guid:
                    # The tool has not changed between the compared changeset revisions.
                    continue
                if tool_dict["id"] == old_id and tool_dict["version"] != version:
                    # The tool version is different, so we've found the parent.
                    return tool_dict["guid"]
        if parent_id is None:
            # The tool did not change through all of the changeset revisions.
            return old_id

    def get_query_for_setting_metadata_on_repositories(self, my_writable=False, order=True):
        """
        Return a query containing repositories for resetting metadata.  The order parameter
        is used for displaying the list of repositories ordered alphabetically for display on
        a page.  When called from the Tool Shed API, order is False.
        """
        # When called from the Tool Shed API, the metadata is reset on all repositories of types
        # repository_suite_definition and tool_dependency_definition in addition to other selected
        # repositories.
        if my_writable:
            username = self.user.username
            clause_list = []
            for repository in self.sa_session.query(self.app.model.Repository).filter(
                self.app.model.Repository.table.c.deleted == false()
            ):
                # Always reset metadata on all repositories of types repository_suite_definition and
                # tool_dependency_definition.
                if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION, rt_util.TOOL_DEPENDENCY_DEFINITION]:
                    clause_list.append(self.app.model.Repository.table.c.id == repository.id)
                else:
                    allow_push = repository.allow_push()
                    if allow_push:
                        # Include all repositories that are writable by the current user.
                        allow_push_usernames = allow_push.split(",")
                        if username in allow_push_usernames:
                            clause_list.append(self.app.model.Repository.table.c.id == repository.id)
            if clause_list:
                if order:
                    return (
                        self.sa_session.query(self.app.model.Repository)
                        .filter(or_(*clause_list))
                        .order_by(self.app.model.Repository.table.c.name, self.app.model.Repository.table.c.user_id)
                    )
                else:
                    return self.sa_session.query(self.app.model.Repository).filter(or_(*clause_list))
            else:
                # Return an empty query.
                return self.sa_session.query(self.app.model.Repository).filter(
                    self.app.model.Repository.table.c.id == -1
                )
        else:
            if order:
                return (
                    self.sa_session.query(self.app.model.Repository)
                    .filter(self.app.model.Repository.table.c.deleted == false())
                    .order_by(self.app.model.Repository.table.c.name, self.app.model.Repository.table.c.user_id)
                )
            else:
                return self.sa_session.query(self.app.model.Repository).filter(
                    self.app.model.Repository.table.c.deleted == false()
                )

    def new_datatypes_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each datatype in the repository with the new metadata
        in self.metadata_dict to determine if a new repository_metadata table record is required
        or if the last saved metadata record can be updated for datatypes instead.
        """
        # Datatypes are stored in metadata as a list of dictionaries that looks like:
        # [{'dtype': 'galaxy.datatypes.data:Text', 'subclass': 'True', 'extension': 'acedb'}]
        if "datatypes" in self.metadata_dict:
            current_datatypes = self.metadata_dict["datatypes"]
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "datatypes" in metadata:
                        ancestor_datatypes = metadata["datatypes"]
                        # The saved metadata must be a subset of the new metadata.
                        datatype_comparison = self.compare_datatypes(ancestor_datatypes, current_datatypes)
                        if datatype_comparison == self.NOT_EQUAL_AND_NOT_SUBSET:
                            return True
                        else:
                            return False
                    else:
                        # The new metadata includes datatypes, but the stored metadata does not,
                        # so we can update the stored metadata.
                        return False
                else:
                    # There is no stored metadata, so we can update the metadata column in the
                    # repository_metadata table.
                    return False
            else:
                # There is no stored repository metadata, so we need to create a new repository_metadata
                # table record.
                return True
        # self.metadata_dict includes no metadata for datatypes, so a new repository_metadata
        # table record is not needed.
        return False

    def new_metadata_required_for_utilities(self):
        """
        This method compares the last stored repository_metadata record associated with self.repository
        against the contents of self.metadata_dict and returns True or False for the union set of Galaxy
        utilities contained in both metadata dictionaries.  The metadata contained in self.metadata_dict
        may not be a subset of that contained in the last stored repository_metadata record associated with
        self.repository because one or more Galaxy utilities may have been deleted from self.repository in
        the new tip.
        """
        repository_metadata = metadata_util.get_latest_repository_metadata(
            self.app, self.repository.id, downloadable=False
        )
        datatypes_required = self.new_datatypes_metadata_required(repository_metadata)
        # Uncomment the following if we decide that README files should affect how installable
        # repository revisions are defined.  See the NOTE in the compare_readme_files() method.
        # readme_files_required = sewlf.new_readme_files_metadata_required( repository_metadata )
        repository_dependencies_required = self.new_repository_dependency_metadata_required(repository_metadata)
        tools_required = self.new_tool_metadata_required(repository_metadata)
        tool_dependencies_required = self.new_tool_dependency_metadata_required(repository_metadata)
        workflows_required = self.new_workflow_metadata_required(repository_metadata)
        if (
            datatypes_required
            or repository_dependencies_required
            or tools_required
            or tool_dependencies_required
            or workflows_required
        ):
            return True
        return False

    def new_readme_files_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each readme file in the repository with the new metadata
        in self.metadata_dict to determine if a new repository_metadata table record is required or
        if the last saved metadata record can be updated for readme files instead.
        """
        # Repository README files are kind of a special case because they have no effect on reproducibility.
        # We'll simply inspect the file names to determine if any that exist in the saved metadata are
        # eliminated from the new metadata in self.metadata_dict.
        if "readme_files" in self.metadata_dict:
            current_readme_files = self.metadata_dict["readme_files"]
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "readme_files" in metadata:
                        ancestor_readme_files = metadata["readme_files"]
                        # The saved metadata must be a subset of the new metadata.
                        readme_file_comparison = self.compare_readme_files(ancestor_readme_files, current_readme_files)
                        if readme_file_comparison == self.NOT_EQUAL_AND_NOT_SUBSET:
                            return True
                        else:
                            return False
                    else:
                        # The new metadata includes readme_files, but the stored metadata does not, so
                        # we can update the stored metadata.
                        return False
                else:
                    # There is no stored metadata, so we can update the metadata column in the repository_metadata
                    # table.
                    return False
            else:
                # There is no stored repository metadata, so we need to create a new repository_metadata
                # table record.
                return True
        # self.metadata_dict includes no metadata for readme_files, so a new repository_metadata
        # table record is not needed.
        return False

    def new_repository_dependency_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each repository dependency in the repository
        with the new metadata in self.metadata_dict to determine if a new repository_metadata
        table record is required or if the last saved metadata record can be updated for
        repository_dependencies instead.
        """
        if repository_metadata:
            metadata = repository_metadata.metadata
            if "repository_dependencies" in metadata:
                saved_repository_dependencies = metadata["repository_dependencies"]["repository_dependencies"]
                new_repository_dependencies_metadata = self.metadata_dict.get("repository_dependencies", None)
                if new_repository_dependencies_metadata:
                    new_repository_dependencies = self.metadata_dict["repository_dependencies"][
                        "repository_dependencies"
                    ]
                    # TODO: We used to include the following here to handle the case where repository
                    # dependency definitions were deleted.  However this erroneously returned True in
                    # cases where is should not have done so.  This usually occurred where multiple single
                    # files were uploaded when a single tarball should have been.  We need to implement
                    # support for handling deleted repository dependency definitions so that we can guarantee
                    # reproducibility, but we need to do it in a way that is better than the following.
                    # for new_repository_dependency in new_repository_dependencies:
                    #     if new_repository_dependency not in saved_repository_dependencies:
                    #         return True
                    # The saved metadata must be a subset of the new metadata.
                    for saved_repository_dependency in saved_repository_dependencies:
                        if saved_repository_dependency not in new_repository_dependencies:
                            # In some cases, the only difference between a dependency definition in the lists
                            # is the changeset_revision value.  We'll check to see if this is the case, and if
                            # the defined dependency is a repository that has metadata set only on its tip.
                            if not self.different_revision_defines_tip_only_repository_dependency(
                                saved_repository_dependency, new_repository_dependencies
                            ):
                                return True
                    return False
                else:
                    # The repository_dependencies.xml file must have been deleted, so create a new
                    # repository_metadata record so we always have access to the deleted file.
                    return True
            else:
                return False
        else:
            if "repository_dependencies" in self.metadata_dict:
                # There is no saved repository metadata, so we need to create a new repository_metadata record.
                return True
            else:
                # self.metadata_dict includes no metadata for repository dependencies, so a new repository_metadata
                # record is not needed.
                return False

    def new_tool_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each tool in the repository with the new metadata in
        self.metadata_dict to determine if a new repository_metadata table record is required, or if
        the last saved metadata record can be updated instead.
        """
        if "tools" in self.metadata_dict:
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "tools" in metadata:
                        saved_tool_ids = []
                        # The metadata for one or more tools was successfully generated in the past
                        # for this repository, so we first compare the version string for each tool id
                        # in self.metadata_dict with what was previously saved to see if we need to create
                        # a new table record or if we can simply update the existing record.
                        for new_tool_metadata_dict in self.metadata_dict["tools"]:
                            for saved_tool_metadata_dict in metadata["tools"]:
                                if saved_tool_metadata_dict["id"] not in saved_tool_ids:
                                    saved_tool_ids.append(saved_tool_metadata_dict["id"])
                                if new_tool_metadata_dict["id"] == saved_tool_metadata_dict["id"]:
                                    if new_tool_metadata_dict["version"] != saved_tool_metadata_dict["version"]:
                                        return True
                        # So far, a new metadata record is not required, but we still have to check to see if
                        # any new tool ids exist in self.metadata_dict that are not in the saved metadata.  We do
                        # this because if a new tarball was uploaded to a repository that included tools, it
                        # may have removed existing tool files if they were not included in the uploaded tarball.
                        for new_tool_metadata_dict in self.metadata_dict["tools"]:
                            if new_tool_metadata_dict["id"] not in saved_tool_ids:
                                return True
                        return False
                    else:
                        # The new metadata includes tools, but the stored metadata does not, so we can
                        # update the stored metadata.
                        return False
                else:
                    # There is no stored metadata, so we can update the metadata column in the
                    # repository_metadata table.
                    return False
            else:
                # There is no stored repository metadata, so we need to create a new repository_metadata
                # table record.
                return True
        # self.metadata_dict includes no metadata for tools, so a new repository_metadata table
        # record is not needed.
        return False

    def new_tool_dependency_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each tool dependency in the repository with the new
        metadata in self.metadata_dict to determine if a new repository_metadata table record is
        required or if the last saved metadata record can be updated for tool_dependencies instead.
        """
        if repository_metadata:
            metadata = repository_metadata.metadata
            if metadata:
                if "tool_dependencies" in metadata:
                    saved_tool_dependencies = metadata["tool_dependencies"]
                    new_tool_dependencies = self.metadata_dict.get("tool_dependencies", None)
                    if new_tool_dependencies:
                        # TODO: We used to include the following here to handle the case where
                        # tool dependency definitions were deleted.  However, this erroneously
                        # returned True in cases where is should not have done so.  This usually
                        # occurred where multiple single files were uploaded when a single tarball
                        # should have been.  We need to implement support for handling deleted
                        # tool dependency definitions so that we can guarantee reproducibility,
                        # but we need to do it in a way that is better than the following.
                        # for new_tool_dependency in new_tool_dependencies:
                        #     if new_tool_dependency not in saved_tool_dependencies:
                        #         return True
                        # The saved metadata must be a subset of the new metadata.
                        for saved_tool_dependency in saved_tool_dependencies:
                            if saved_tool_dependency not in new_tool_dependencies:
                                return True
                        return False
                    else:
                        # The tool_dependencies.xml file must have been deleted, so create a new
                        # repository_metadata record so we always have
                        # access to the deleted file.
                        return True
                else:
                    return False
            else:
                # We have repository metadata that does not include metadata for any tool dependencies
                # in the repository, so we can update the existing repository metadata.
                return False
        else:
            if "tool_dependencies" in self.metadata_dict:
                # There is no saved repository metadata, so we need to create a new repository_metadata
                # record.
                return True
            else:
                # self.metadata_dict includes no metadata for tool dependencies, so a new repository_metadata
                # record is not needed.
                return False

    def new_workflow_metadata_required(self, repository_metadata):
        """
        Currently everything about an exported workflow except the name is hard-coded, so
        there's no real way to differentiate versions of exported workflows.  If this changes
        at some future time, this method should be enhanced accordingly.
        """
        if "workflows" in self.metadata_dict:
            if repository_metadata:
                # The repository has metadata, so update the workflows value -
                # no new record is needed.
                return False
            else:
                # There is no saved repository metadata, so we need to create a
                # new repository_metadata table record.
                return True
        # self.metadata_dict includes no metadata for workflows, so a new
        # repository_metadata table record is not needed.
        return False

    def reset_all_metadata_on_repository_in_tool_shed(self):
        """Reset all metadata on a single repository in a tool shed."""
        log.debug(f"Resetting all metadata on repository: {self.repository.name}")
        repo = self.repository.hg_repo
        # The list of changeset_revisions refers to repository_metadata records that have been created
        # or updated.  When the following loop completes, we'll delete all repository_metadata records
        # for this repository that do not have a changeset_revision value in this list.
        changeset_revisions = []
        # When a new repository_metadata record is created, it always uses the values of
        # metadata_changeset_revision and metadata_dict.
        metadata_changeset_revision = None
        metadata_dict = None
        ancestor_changeset_revision = None
        ancestor_metadata_dict = None
        for changeset in self.repository.get_changesets_for_setting_metadata(self.app):
            work_dir = tempfile.mkdtemp(prefix="tmp-toolshed-ramorits")
            ctx = repo[changeset]
            log.debug("Cloning repository changeset revision: %s", str(ctx.rev()))
            cloned_ok, error_message = hg_util.clone_repository(self.repository_clone_url, work_dir, str(ctx.rev()))
            if cloned_ok:
                log.debug("Generating metadata for changeset revision: %s", str(ctx.rev()))
                self.set_changeset_revision(str(ctx))
                self.set_repository_files_dir(work_dir)
                self.generate_metadata_for_changeset_revision()
                if self.metadata_dict:
                    if metadata_changeset_revision is None and metadata_dict is None:
                        # We're at the first change set in the change log.
                        metadata_changeset_revision = self.changeset_revision
                        metadata_dict = self.metadata_dict
                    if ancestor_changeset_revision:
                        # Compare metadata from ancestor and current.  The value of comparison will be one of:
                        # self.NO_METADATA - no metadata for either ancestor or current, so continue from current
                        # self.EQUAL - ancestor metadata is equivalent to current metadata, so continue from current
                        # self.SUBSET - ancestor metadata is a subset of current metadata, so continue from current
                        # self.NOT_EQUAL_AND_NOT_SUBSET - ancestor metadata is neither equal to nor a subset of current
                        # metadata, so persist ancestor metadata.
                        comparison = self.compare_changeset_revisions(
                            ancestor_changeset_revision, ancestor_metadata_dict
                        )
                        if comparison in [self.NO_METADATA, self.EQUAL, self.SUBSET]:
                            ancestor_changeset_revision = self.changeset_revision
                            ancestor_metadata_dict = self.metadata_dict
                        elif comparison == self.NOT_EQUAL_AND_NOT_SUBSET:
                            metadata_changeset_revision = ancestor_changeset_revision
                            metadata_dict = ancestor_metadata_dict
                            self.create_or_update_repository_metadata(metadata_changeset_revision, metadata_dict)
                            changeset_revisions.append(metadata_changeset_revision)
                            ancestor_changeset_revision = self.changeset_revision
                            ancestor_metadata_dict = self.metadata_dict
                    else:
                        # We're at the beginning of the change log.
                        ancestor_changeset_revision = self.changeset_revision
                        ancestor_metadata_dict = self.metadata_dict
                    if not ctx.children():
                        metadata_changeset_revision = self.changeset_revision
                        metadata_dict = self.metadata_dict
                        # We're at the end of the change log.
                        self.create_or_update_repository_metadata(metadata_changeset_revision, metadata_dict)
                        changeset_revisions.append(metadata_changeset_revision)
                        ancestor_changeset_revision = None
                        ancestor_metadata_dict = None
                elif ancestor_metadata_dict:
                    # We reach here only if self.metadata_dict is empty and ancestor_metadata_dict is not.
                    if not ctx.children():
                        # We're at the end of the change log.
                        self.create_or_update_repository_metadata(metadata_changeset_revision, metadata_dict)
                        changeset_revisions.append(metadata_changeset_revision)
                        ancestor_changeset_revision = None
                        ancestor_metadata_dict = None
            basic_util.remove_dir(work_dir)
        # Delete all repository_metadata records for this repository that do not have a changeset_revision
        # value in changeset_revisions.
        self.clean_repository_metadata(changeset_revisions)
        # Set tool version information for all downloadable changeset revisions.  Get the list of changeset
        # revisions from the changelog.
        self.reset_all_tool_versions(repo)

    def reset_all_tool_versions(self, repo):
        """Reset tool version lineage for those changeset revisions that include valid tools."""
        encoded_repository_id = self.app.security.encode_id(self.repository.id)
        changeset_revisions_that_contain_tools = []
        for changeset in repo.changelog:
            changeset_revision = str(repo[changeset])
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                self.app, encoded_repository_id, changeset_revision
            )
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if metadata.get("tools", None):
                        changeset_revisions_that_contain_tools.append(changeset_revision)
        # The list of changeset_revisions_that_contain_tools is now filtered to contain only those that
        # are downloadable and contain tools.  If a repository includes tools, build a dictionary of
        # { 'tool id' : 'parent tool id' } pairs for each tool in each changeset revision.
        for index, changeset_revision in enumerate(changeset_revisions_that_contain_tools):
            tool_versions_dict = {}
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                self.app, encoded_repository_id, changeset_revision
            )
            assert repository_metadata
            metadata = repository_metadata.metadata
            tool_dicts = metadata["tools"]
            if index == 0:
                # The first changeset_revision is a special case because it will have no ancestor
                # changeset_revisions in which to match tools.  The parent tool id for tools in the
                # first changeset_revision will be the "old_id" in the tool config.
                for tool_dict in tool_dicts:
                    tool_versions_dict[tool_dict["guid"]] = tool_dict["id"]
            else:
                for tool_dict in tool_dicts:
                    parent_id = self.get_parent_id(
                        encoded_repository_id,
                        tool_dict["id"],
                        tool_dict["version"],
                        tool_dict["guid"],
                        changeset_revisions_that_contain_tools[0:index],
                    )
                    tool_versions_dict[tool_dict["guid"]] = parent_id
            if tool_versions_dict:
                repository_metadata.tool_versions = tool_versions_dict
                self.sa_session.add(repository_metadata)
                session = self.sa_session()
                with transaction(session):
                    session.commit()

    def reset_metadata_on_selected_repositories(self, **kwd):
        """
        Inspect the repository changelog to reset metadata for all appropriate changeset revisions.
        This method is called from both Galaxy and the Tool Shed.
        """
        repository_ids = util.listify(kwd.get("repository_ids", None))
        message = ""
        status = "done"
        if repository_ids:
            successful_count = 0
            unsuccessful_count = 0
            for repository_id in repository_ids:
                try:
                    repository = repository_util.get_repository_in_tool_shed(self.app, repository_id)
                    self.set_repository(repository)
                    self.resetting_all_metadata_on_repository = True
                    self.reset_all_metadata_on_repository_in_tool_shed()
                    if self.invalid_file_tups:
                        message = tool_util.generate_message_for_invalid_tools(
                            self.app, self.invalid_file_tups, repository, None, as_html=False
                        )
                        log.debug(message)
                        unsuccessful_count += 1
                    else:
                        log.debug(
                            "Successfully reset metadata on repository %s owned by %s",
                            repository.name,
                            repository.user.username,
                        )
                        successful_count += 1
                except Exception:
                    log.exception("Error attempting to reset metadata on repository %s", str(repository.name))
                    unsuccessful_count += 1
            message = "Successfully reset metadata on %d %s.  " % (
                successful_count,
                inflector.cond_plural(successful_count, "repository"),
            )
            if unsuccessful_count:
                message += "Error setting metadata on %d %s - see the paster log for details.  " % (
                    unsuccessful_count,
                    inflector.cond_plural(unsuccessful_count, "repository"),
                )
        else:
            message = "Select at least one repository to on which to reset all metadata."
            status = "error"
        return message, status

    def set_repository(self, repository):
        super().set_repository(repository)
        self.repository_clone_url = common_util.generate_clone_url_for_repository_in_tool_shed(self.user, repository)

    def set_repository_metadata(self, host, content_alert_str="", **kwd):
        """
        Set metadata using the self.repository's current disk files, returning specific error
        messages (if any) to alert the repository owner that the changeset has problems.
        """
        message = ""
        status = "done"
        encoded_id = self.app.security.encode_id(self.repository.id)
        repo = self.repository.hg_repo
        self.generate_metadata_for_changeset_revision()
        if self.metadata_dict:
            repository_metadata = None
            repository_type_class = self.app.repository_types_registry.get_class_by_label(self.repository.type)
            tip_only = isinstance(repository_type_class, TipOnly)
            if not tip_only and self.new_metadata_required_for_utilities():
                # Create a new repository_metadata table row.
                repository_metadata = self.create_or_update_repository_metadata(
                    self.repository.tip(), self.metadata_dict
                )
                # If this is the first record stored for this repository, see if we need to send any email alerts.
                if len(self.repository.downloadable_revisions) == 1:
                    suc.handle_email_alerts(
                        self.app, host, self.repository, content_alert_str="", new_repo_alert=True, admin_only=False
                    )
            else:
                # Update the latest stored repository metadata with the contents and attributes of self.metadata_dict.
                repository_metadata = metadata_util.get_latest_repository_metadata(
                    self.app, self.repository.id, downloadable=False
                )
                if repository_metadata:
                    downloadable = metadata_util.is_downloadable(self.metadata_dict)
                    # Update the last saved repository_metadata table row.
                    repository_metadata.changeset_revision = self.repository.tip()
                    repository_metadata.metadata = self.metadata_dict
                    repository_metadata.downloadable = downloadable
                    repository_metadata.includes_datatypes = False
                    # We don't store information about the special type of repository dependency that is needed only for
                    # compiling a tool dependency defined for the dependent repository.
                    repository_dependencies_dict = self.metadata_dict.get("repository_dependencies", {})
                    repository_dependencies = repository_dependencies_dict.get("repository_dependencies", [])
                    (
                        has_repository_dependencies,
                        has_repository_dependencies_only_if_compiling_contained_td,
                    ) = repository_util.get_repository_dependency_types(repository_dependencies)
                    repository_metadata.has_repository_dependencies = has_repository_dependencies
                    if "tool_dependencies" in self.metadata_dict:
                        repository_metadata.includes_tool_dependencies = True
                    else:
                        repository_metadata.includes_tool_dependencies = False
                    if "tools" in self.metadata_dict:
                        repository_metadata.includes_tools = True
                    else:
                        repository_metadata.includes_tools = False
                    repository_metadata.includes_workflows = False
                    repository_metadata.missing_test_components = False
                    self.sa_session.add(repository_metadata)
                    session = self.sa_session()
                    with transaction(session):
                        session.commit()
                else:
                    # There are no metadata records associated with the repository.
                    repository_metadata = self.create_or_update_repository_metadata(
                        self.repository.tip(), self.metadata_dict
                    )
            if "tools" in self.metadata_dict and repository_metadata and status != "error":
                # Set tool versions on the new downloadable change set.  The order of the list of changesets is
                # critical, so we use the repo's changelog.
                changeset_revisions = []
                for changeset in repo.changelog:
                    changeset_revision = str(repo[changeset])
                    if metadata_util.get_repository_metadata_by_changeset_revision(
                        self.app, encoded_id, changeset_revision
                    ):
                        changeset_revisions.append(changeset_revision)
                self.add_tool_versions(encoded_id, repository_metadata, changeset_revisions)
        elif len(repo) == 1 and not self.invalid_file_tups:
            message = "Revision <b>%s</b> includes no Galaxy utilities for which metadata can " % str(
                self.repository.tip()
            )
            message += "be defined so this revision cannot be automatically installed into a local Galaxy instance."
            status = "error"
        if self.invalid_file_tups:
            message = tool_util.generate_message_for_invalid_tools(
                self.app, self.invalid_file_tups, self.repository, self.metadata_dict
            )
            status = "error"
        return message, status

    def set_repository_metadata_due_to_new_tip(self, host, content_alert_str=None, **kwd):
        """Set metadata on the tip of self.repository in the tool shed."""
        error_message, status = self.set_repository_metadata(host, content_alert_str=content_alert_str, **kwd)
        return status, error_message



import os
import shutil

import tool_shed.repository_types.util as rt_util
from tool_shed.util import (
    commit_util,
    xml_util,
)


def upload_tar(
    trans,
    rdah,
    tdah,
    repository,
    tar,
    uploaded_file,
    upload_point,
    remove_repo_files_not_in_tar,
    commit_message,
    new_repo_alert,
):
    # Upload a tar archive of files.
    undesirable_dirs_removed = 0
    undesirable_files_removed = 0
    check_results = commit_util.check_archive(repository, tar)
    if check_results.invalid:
        tar.close()
        uploaded_file.close()
        message = "{} Invalid paths were: {}".format(" ".join(check_results.errors), ", ".join(check_results.invalid))
        return False, message, [], "", undesirable_dirs_removed, undesirable_files_removed
    else:
        repo_dir = repository.repo_path(trans.app)
        if upload_point is not None:
            full_path = os.path.abspath(os.path.join(repo_dir, upload_point))
        else:
            full_path = os.path.abspath(repo_dir)
        undesirable_files_removed = len(check_results.undesirable_files)
        undesirable_dirs_removed = len(check_results.undesirable_dirs)
        filenames_in_archive = [ti.name for ti in check_results.valid]
        # Extract the uploaded tar to the load_point within the repository hierarchy.
        tar.extractall(path=full_path, members=check_results.valid)
        tar.close()
        uploaded_file.close()
        for filename in filenames_in_archive:
            uploaded_file_name = os.path.join(full_path, filename)
            if os.path.split(uploaded_file_name)[-1] == rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME:
                # Inspect the contents of the file to see if toolshed or changeset_revision attributes
                # are missing and if so, set them appropriately.
                altered, root_elem, error_message = rdah.handle_tag_attributes(uploaded_file_name)
                if error_message:
                    return False, error_message, [], "", [], []
                elif altered:
                    tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                    shutil.move(tmp_filename, uploaded_file_name)
            elif os.path.split(uploaded_file_name)[-1] == rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME:
                # Inspect the contents of the file to see if toolshed or changeset_revision
                # attributes are missing and if so, set them appropriately.
                altered, root_elem, error_message = tdah.handle_tag_attributes(uploaded_file_name)
                if error_message:
                    return False, error_message, [], "", [], []
                if altered:
                    tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                    shutil.move(tmp_filename, uploaded_file_name)
        return commit_util.handle_directory_changes(
            trans.app,
            trans.request.host,
            trans.user.username,
            repository,
            full_path,
            filenames_in_archive,
            remove_repo_files_not_in_tar,
            new_repo_alert,
            commit_message,
            undesirable_dirs_removed,
            undesirable_files_removed,
        )



"""
API operations on Group objects.
"""
import logging
from typing import (
    Any,
    Dict,
)

from galaxy.managers.context import ProvidesAppContext
from galaxy.managers.groups import GroupsManager
from galaxy.schema.fields import EncodedDatabaseIdField
from galaxy.web import (
    expose_api,
    require_admin,
)
from galaxy.webapps.galaxy.api import (
    BaseGalaxyAPIController,
    depends,
)

log = logging.getLogger(__name__)


class GroupAPIController(BaseGalaxyAPIController):
    manager = depends(GroupsManager)

    @expose_api
    @require_admin
    def index(self, trans: ProvidesAppContext, **kwd):
        """
        GET /api/groups
        Displays a collection (list) of groups.
        """
        return self.manager.index(trans)

    @expose_api
    @require_admin
    def create(self, trans: ProvidesAppContext, payload: Dict[str, Any], **kwd):
        """
        POST /api/groups
        Creates a new group.
        """
        return self.manager.create(trans, payload)

    @expose_api
    @require_admin
    def show(self, trans: ProvidesAppContext, id: EncodedDatabaseIdField, **kwd):
        """
        GET /api/groups/{encoded_group_id}
        Displays information about a group.
        """
        return self.manager.show(trans, EncodedDatabaseIdField.decode(id))

    @expose_api
    @require_admin
    def update(self, trans: ProvidesAppContext, id: EncodedDatabaseIdField, payload: Dict[str, Any], **kwd):
        """
        PUT /api/groups/{encoded_group_id}
        Modifies a group.
        """
        self.manager.update(trans, EncodedDatabaseIdField.decode(id), payload)



import logging
from json import loads

from markupsafe import escape
from paste.httpexceptions import (
    HTTPBadRequest,
    HTTPNotFound,
)
from sqlalchemy import (
    desc,
    false,
    or_,
    text,
    true,
)
from sqlalchemy.orm import (
    joinedload,
    undefer,
)

from galaxy import (
    model,
    util,
    web,
)
from galaxy.managers.hdas import HDAManager
from galaxy.managers.sharable import SlugBuilder
from galaxy.model.base import transaction
from galaxy.model.item_attrs import (
    UsesAnnotations,
    UsesItemRatings,
)
from galaxy.structured_app import StructuredApp
from galaxy.util import (
    sanitize_text,
    unicodify,
)
from galaxy.util.sanitize_html import sanitize_html
from galaxy.visualization.data_providers.phyloviz import PhylovizDataProvider
from galaxy.visualization.genomes import (
    decode_dbkey,
    GenomeRegion,
)
from galaxy.visualization.plugins import registry
from galaxy.web.framework.helpers import (
    grids,
    time_ago,
)
from galaxy.webapps.base.controller import (
    BaseUIController,
    SharableMixin,
    UsesVisualizationMixin,
)
from ..api import depends

log = logging.getLogger(__name__)


#
# -- Grids --
#
class HistoryDatasetsSelectionGrid(grids.Grid):
    class DbKeyColumn(grids.GridColumn):
        def filter(self, trans, user, query, dbkey):
            """Filter by dbkey through a raw SQL b/c metadata is a BLOB."""
            dbkey_user, dbkey = decode_dbkey(dbkey)
            dbkey = dbkey.replace("'", "\\'")
            return query.filter(
                or_(text(f'metadata like \'%"dbkey": ["{dbkey}"]%\'', f'metadata like \'%"dbkey": "{dbkey}"%\''))
            )

    class HistoryColumn(grids.GridColumn):
        def get_value(self, trans, grid, hda):
            return escape(hda.history.name)

        def sort(self, trans, query, ascending, column_name=None):
            """Sort query using this column."""
            return grids.GridColumn.sort(self, trans, query, ascending, column_name="history_id")

    available_tracks = None
    title = "Add Datasets"
    model_class = model.HistoryDatasetAssociation
    default_filter = {"deleted": "False", "shared": "All"}
    default_sort_key = "-hid"
    columns = [
        grids.GridColumn("Id", key="hid"),
        grids.TextColumn("Name", key="name", model_class=model.HistoryDatasetAssociation),
        grids.TextColumn("Type", key="extension", model_class=model.HistoryDatasetAssociation),
        grids.TextColumn("history_id", key="history_id", model_class=model.HistoryDatasetAssociation, visible=False),
        HistoryColumn("History", key="history", visible=True),
        DbKeyColumn("Build", key="dbkey", model_class=model.HistoryDatasetAssociation, visible=True, sortable=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search name and filetype",
            cols_to_filter=[columns[1], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwargs):
        return trans.sa_session.query(self.model_class).join(model.History.table).join(model.Dataset.table)

    def apply_query_filter(self, trans, query, **kwargs):
        if self.available_tracks is None:
            self.available_tracks = trans.app.datatypes_registry.get_available_tracks()
        return (
            query.filter(model.History.user == trans.user)
            .filter(model.HistoryDatasetAssociation.extension.in_(self.available_tracks))
            .filter(model.Dataset.state == model.Dataset.states.OK)
            .filter(model.HistoryDatasetAssociation.deleted == false())
            .filter(model.HistoryDatasetAssociation.visible == true())
        )


class LibraryDatasetsSelectionGrid(grids.Grid):
    available_tracks = None
    title = "Add Datasets"
    model_class = model.LibraryDatasetDatasetAssociation
    default_filter = {"deleted": "False"}
    default_sort_key = "-id"
    columns = [
        grids.GridColumn("Id", key="id"),
        grids.TextColumn("Name", key="name", model_class=model.LibraryDatasetDatasetAssociation),
        grids.TextColumn("Type", key="extension", model_class=model.LibraryDatasetDatasetAssociation),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search name and filetype",
            cols_to_filter=[columns[1], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwargs):
        return trans.sa_session.query(self.model_class).join(model.Dataset.table)

    def apply_query_filter(self, trans, query, **kwargs):
        if self.available_tracks is None:
            self.available_tracks = trans.app.datatypes_registry.get_available_tracks()
        return (
            query.filter(model.LibraryDatasetDatasetAssociation.user == trans.user)
            .filter(model.LibraryDatasetDatasetAssociation.extension.in_(self.available_tracks))
            .filter(model.Dataset.state == model.Dataset.states.OK)
            .filter(model.LibraryDatasetDatasetAssociation.deleted == false())
            .filter(model.LibraryDatasetDatasetAssociation.visible == true())
        )


class TracksterSelectionGrid(grids.Grid):
    title = "Insert into visualization"
    model_class = model.Visualization
    default_sort_key = "-update_time"
    use_paging = False
    show_item_checkboxes = True
    columns = [
        grids.TextColumn("Title", key="title", model_class=model.Visualization, filterable="standard"),
        grids.TextColumn("Build", key="dbkey", model_class=model.Visualization),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
    ]

    def build_initial_query(self, trans, **kwargs):
        return trans.sa_session.query(self.model_class)

    def apply_query_filter(self, trans, query, **kwargs):
        return (
            query.filter(self.model_class.user_id == trans.user.id)
            .filter(self.model_class.deleted == false())
            .filter(self.model_class.type == "trackster")
        )


class VisualizationListGrid(grids.Grid):
    def get_url_args(item):
        """
        Returns dictionary used to create item link.
        """
        url_kwargs = dict(controller="visualization", id=item.id)
        # TODO: hack to build link to saved visualization - need trans in this function instead in order to do
        # link_data = trans.app.visualizations_registry.get_visualizations( trans, item )
        if item.type in registry.VisualizationsRegistry.BUILT_IN_VISUALIZATIONS:
            url_kwargs["action"] = item.type
        else:
            url_kwargs["__route_name__"] = "saved_visualization"
            url_kwargs["visualization_name"] = item.type
            url_kwargs["action"] = "saved"
        return url_kwargs

    def get_display_name(self, trans, item):
        if trans.app.visualizations_registry and item.type in trans.app.visualizations_registry.plugins:
            plugin = trans.app.visualizations_registry.plugins[item.type]
            return plugin.config.get("name", item.type)
        return item.type

    # Grid definition
    title = "Saved Visualizations"
    model_class = model.Visualization
    default_sort_key = "-update_time"
    default_filter = dict(title="All", deleted="False", tags="All", sharing="All")
    columns = [
        grids.TextColumn("Title", key="title", attach_popup=True, link=get_url_args),
        grids.TextColumn("Type", method="get_display_name"),
        grids.TextColumn("Build", key="dbkey"),
        grids.IndividualTagsColumn(
            "Tags",
            key="tags",
            model_tag_association_class=model.VisualizationTagAssociation,
            filterable="advanced",
            grid_name="VisualizationListGrid",
        ),
        grids.SharingStatusColumn("Sharing", key="sharing", filterable="advanced", sortable=False),
        grids.GridColumn("Created", key="create_time", format=time_ago),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    operations = [
        grids.GridOperation("Open", allow_multiple=False, url_args=get_url_args),
        grids.GridOperation(
            "Edit Attributes", allow_multiple=False, url_args=dict(controller="", action="visualizations/edit")
        ),
        grids.GridOperation("Copy", allow_multiple=False, condition=(lambda item: not item.deleted)),
        grids.GridOperation(
            "Share or Publish",
            allow_multiple=False,
            condition=(lambda item: not item.deleted),
            url_args=dict(controller="", action="visualizations/sharing"),
        ),
        grids.GridOperation(
            "Delete",
            condition=(lambda item: not item.deleted),
            confirm="Are you sure you want to delete this visualization?",
        ),
    ]

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter_by(user=trans.user, deleted=False)


class VisualizationAllPublishedGrid(grids.Grid):
    # Grid definition
    use_panels = True
    title = "Published Visualizations"
    model_class = model.Visualization
    default_sort_key = "update_time"
    default_filter = dict(title="All", username="All")
    columns = [
        grids.PublicURLColumn("Title", key="title", filterable="advanced"),
        grids.OwnerAnnotationColumn(
            "Annotation",
            key="annotation",
            model_annotation_association_class=model.VisualizationAnnotationAssociation,
            filterable="advanced",
        ),
        grids.OwnerColumn("Owner", key="username", model_class=model.User, filterable="advanced"),
        grids.CommunityRatingColumn("Community Rating", key="rating"),
        grids.CommunityTagsColumn(
            "Community Tags",
            key="tags",
            model_tag_association_class=model.VisualizationTagAssociation,
            filterable="advanced",
            grid_name="VisualizationAllPublishedGrid",
        ),
        grids.ReverseSortColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search title, annotation, owner, and tags",
            cols_to_filter=[columns[0], columns[1], columns[2], columns[4]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwargs):
        # See optimization description comments and TODO for tags in matching public histories query.
        return (
            trans.sa_session.query(self.model_class)
            .join(self.model_class.user)
            .options(
                joinedload(self.model_class.user).load_only("username"),
                joinedload(self.model_class.annotations),
                undefer("average_rating"),
            )
        )

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter(self.model_class.deleted == false()).filter(self.model_class.published == true())


class VisualizationController(
    BaseUIController, SharableMixin, UsesVisualizationMixin, UsesAnnotations, UsesItemRatings
):
    _visualization_list_grid = VisualizationListGrid()
    _published_list_grid = VisualizationAllPublishedGrid()
    _history_datasets_grid = HistoryDatasetsSelectionGrid()
    _library_datasets_grid = LibraryDatasetsSelectionGrid()
    _tracks_grid = TracksterSelectionGrid()
    hda_manager: HDAManager = depends(HDAManager)
    slug_builder: SlugBuilder = depends(SlugBuilder)

    def __init__(self, app: StructuredApp):
        super().__init__(app)

    #
    # -- Functions for listing visualizations. --
    #

    @web.expose
    @web.json
    @web.require_login("see all available libraries")
    def list_libraries(self, trans, **kwargs):
        """List all libraries that can be used for selecting datasets."""
        return self._libraries_grid(trans, **kwargs)

    @web.expose
    @web.json
    @web.require_login("see a history's datasets that can added to this visualization")
    def list_history_datasets(self, trans, **kwargs):
        """List a history's datasets that can be added to a visualization."""
        kwargs["show_item_checkboxes"] = "True"
        return self._history_datasets_grid(trans, **kwargs)

    @web.expose
    @web.json
    @web.require_login("see a history's datasets that can added to this visualization")
    def list_library_datasets(self, trans, **kwargs):
        """List a library's datasets that can be added to a visualization."""
        kwargs["show_item_checkboxes"] = "True"
        return self._library_datasets_grid(trans, **kwargs)

    @web.expose
    @web.json
    def list_tracks(self, trans, **kwargs):
        return self._tracks_grid(trans, **kwargs)

    @web.expose
    @web.json
    def list_published(self, trans, *args, **kwargs):
        grid = self._published_list_grid(trans, **kwargs)
        grid["shared_by_others"] = self._get_shared(trans)
        return grid

    @web.legacy_expose_api
    @web.require_login("use Galaxy visualizations", use_panels=True)
    def list(self, trans, **kwargs):
        message = kwargs.get("message")
        status = kwargs.get("status")
        if "operation" in kwargs and "id" in kwargs:
            session = trans.sa_session
            operation = kwargs["operation"].lower()
            ids = util.listify(kwargs["id"])
            for id in ids:
                if operation == "delete":
                    item = self.get_visualization(trans, id)
                    item.deleted = True
                if operation == "copy":
                    self.copy(trans, **kwargs)
            with transaction(session):
                session.commit()
        kwargs["embedded"] = True
        if message and status:
            kwargs["message"] = sanitize_text(message)
            kwargs["status"] = status
        grid = self._visualization_list_grid(trans, **kwargs)
        grid["shared_by_others"] = self._get_shared(trans)
        return grid

    def _get_shared(self, trans):
        """Identify shared visualizations"""
        shared_by_others = (
            trans.sa_session.query(model.VisualizationUserShareAssociation)
            .filter_by(user=trans.get_user())
            .join(model.Visualization.table)
            .filter(model.Visualization.deleted == false())
            .order_by(desc(model.Visualization.update_time))
            .all()
        )
        return [
            {"username": v.visualization.user.username, "slug": v.visualization.slug, "title": v.visualization.title}
            for v in shared_by_others
        ]

    #
    # -- Functions for operating on visualizations. --
    #

    @web.expose
    @web.require_login("use Galaxy visualizations", use_panels=True)
    def index(self, trans, *args, **kwargs):
        """Lists user's saved visualizations."""
        return self.list(trans, *args, **kwargs)

    @web.expose
    @web.require_login()
    def copy(self, trans, id, **kwargs):
        visualization = self.get_visualization(trans, id, check_ownership=False, check_accessible=True)
        user = trans.get_user()
        owner = visualization.user == user
        new_title = f"Copy of '{visualization.title}'"
        if not owner:
            new_title += f" shared by {visualization.user.email}"

        copied_viz = visualization.copy(user=trans.user, title=new_title)

        # Persist
        session = trans.sa_session
        session.add(copied_viz)
        with transaction(session):
            session.commit()

        # Display the management page
        trans.set_message(f'Created new visualization with name "{copied_viz.title}"')
        return

    @web.expose
    @web.require_login("share Galaxy visualizations")
    def imp(self, trans, id, **kwargs):
        """Import a visualization into user's workspace."""
        # Set referer message.
        referer = trans.request.referer
        if referer and not referer.startswith(f"{trans.request.application_url}{web.url_for('/login')}"):
            referer_message = f"<a href='{escape(referer)}'>return to the previous page</a>"
        else:
            referer_message = f"<a href='{web.url_for('/')}'>go to Galaxy's start page</a>"

        # Do import.
        session = trans.sa_session
        visualization = self.get_visualization(trans, id, check_ownership=False, check_accessible=True)
        if visualization.importable is False:
            return trans.show_error_message(
                f"The owner of this visualization has disabled imports via this link.<br>You can {referer_message}",
                use_panels=True,
            )
        elif visualization.deleted:
            return trans.show_error_message(
                f"You can't import this visualization because it has been deleted.<br>You can {referer_message}",
                use_panels=True,
            )
        else:
            # Create imported visualization via copy.
            #   TODO: need to handle custom db keys.

            imported_visualization = visualization.copy(user=trans.user, title=f"imported: {visualization.title}")

            # Persist
            session = trans.sa_session
            session.add(imported_visualization)
            with transaction(session):
                session.commit()

            # Redirect to load galaxy frames.
            return trans.show_ok_message(
                message="""Visualization "{}" has been imported. <br>You can <a href="{}">start using this visualization</a> or {}.""".format(
                    visualization.title, web.url_for("/visualizations/list"), referer_message
                ),
                use_panels=True,
            )

    @web.expose
    def display_by_username_and_slug(self, trans, username, slug, **kwargs):
        """Display visualization based on a username and slug."""

        # Get visualization.
        session = trans.sa_session
        user = session.query(model.User).filter_by(username=username).first()
        visualization = (
            trans.sa_session.query(model.Visualization).filter_by(user=user, slug=slug, deleted=False).first()
        )
        if visualization is None:
            raise web.httpexceptions.HTTPNotFound()

        # Security check raises error if user cannot access visualization.
        self.security_check(trans, visualization, check_ownership=False, check_accessible=True)

        # Encode page identifier.
        visualization_id = trans.security.encode_id(visualization.id)

        # Redirect to client.
        return trans.response.send_redirect(
            web.url_for(
                controller="published",
                action="visualization",
                id=visualization_id,
            )
        )

    @web.json
    def save(self, trans, vis_json=None, type=None, id=None, title=None, dbkey=None, annotation=None, **kwargs):
        """
        Save a visualization; if visualization does not have an ID, a new
        visualization is created. Returns JSON of visualization.
        """
        # Get visualization attributes from kwargs or from config.
        vis_config = loads(vis_json)
        vis_type = type or vis_config["type"]
        vis_id = id or vis_config.get("id", None)
        vis_title = title or vis_config.get("title", None)
        vis_dbkey = dbkey or vis_config.get("dbkey", None)
        vis_annotation = annotation or vis_config.get("annotation", None)
        return self.save_visualization(trans, vis_config, vis_type, vis_id, vis_title, vis_dbkey, vis_annotation)

    @web.legacy_expose_api
    @web.require_login("edit visualizations")
    def edit(self, trans, payload=None, **kwd):
        """
        Edit a visualization's attributes.
        """
        id = kwd.get("id")
        if not id:
            return self.message_exception(trans, "No visualization id received for editing.")
        trans_user = trans.get_user()
        v = self.get_visualization(trans, id, check_ownership=True)
        if trans.request.method == "GET":
            if v.slug is None:
                self.slug_builder.create_item_slug(trans.sa_session, v)
            return {
                "title": "Edit visualization attributes",
                "inputs": [
                    {"name": "title", "label": "Name", "value": v.title},
                    {
                        "name": "slug",
                        "label": "Identifier",
                        "value": v.slug,
                        "help": "A unique identifier that will be used for public links to this visualization. This field can only contain lowercase letters, numbers, and dashes (-).",
                    },
                    {
                        "name": "dbkey",
                        "label": "Build",
                        "type": "select",
                        "optional": True,
                        "value": v.dbkey,
                        "options": trans.app.genomes.get_dbkeys(trans_user, chrom_info=True),
                        "help": "Parameter to associate your visualization with a database key.",
                    },
                    {
                        "name": "annotation",
                        "label": "Annotation",
                        "value": self.get_item_annotation_str(trans.sa_session, trans.user, v),
                        "help": "A description of the visualization. The annotation is shown alongside published visualizations.",
                    },
                ],
            }
        else:
            v_title = payload.get("title")
            v_slug = payload.get("slug")
            v_dbkey = payload.get("dbkey")
            v_annotation = payload.get("annotation")
            if not v_title:
                return self.message_exception(trans, "Please provide a visualization name is required.")
            elif not v_slug:
                return self.message_exception(trans, "Please provide a unique identifier.")
            elif not self._is_valid_slug(v_slug):
                return self.message_exception(
                    trans, "Visualization identifier can only contain lowercase letters, numbers, and dashes (-)."
                )
            elif (
                v_slug != v.slug
                and trans.sa_session.query(model.Visualization)
                .filter_by(user=v.user, slug=v_slug, deleted=False)
                .first()
            ):
                return self.message_exception(trans, "Visualization id must be unique.")
            else:
                v.title = v_title
                v.slug = v_slug
                v.dbkey = v_dbkey
                if v_annotation:
                    v_annotation = sanitize_html(v_annotation)
                    self.add_item_annotation(trans.sa_session, trans_user, v, v_annotation)
                trans.sa_session.add(v)
                with transaction(trans.sa_session):
                    trans.sa_session.commit()
            return {"message": "Attributes of '%s' successfully saved." % v.title, "status": "success"}

    # ------------------------- registry.
    @web.expose
    @web.require_login("use Galaxy visualizations", use_panels=True)
    def render(self, trans, visualization_name, embedded=None, **kwargs):
        """
        Render the appropriate visualization template, parsing the `kwargs`
        into appropriate variables and resources (such as ORM models)
        based on this visualizations `param` data in visualizations_conf.xml.

        URL: /visualization/show/{visualization_name}
        """
        plugin = self._get_plugin_from_registry(trans, visualization_name)
        try:
            return plugin.render(trans=trans, embedded=embedded, **kwargs)
        except Exception as exception:
            self._handle_plugin_error(trans, visualization_name, exception)

    def _get_plugin_from_registry(self, trans, visualization_name):
        """
        Get the named plugin from the registry.
        :raises HTTPNotFound: if registry has been turned off in config.
        :raises HTTPNotFound: if visualization_name isn't a registered plugin.
        """
        if not trans.app.visualizations_registry:
            raise HTTPNotFound("No visualization registry (possibly disabled in galaxy.ini)")
        return trans.app.visualizations_registry.get_plugin(visualization_name)

    def _handle_plugin_error(self, trans, visualization_name, exception):
        """
        Log, raise if debugging; log and show html message if not.
        """
        log.exception("error rendering visualization (%s)", visualization_name)
        if trans.debug:
            raise exception
        return trans.show_error_message(
            "There was an error rendering the visualization. "
            + "Contact your Galaxy administrator if the problem persists."
            + "<br/>Details: "
            + unicodify(exception),
            use_panels=False,
        )

    @web.expose
    @web.require_login("use Galaxy visualizations", use_panels=True)
    def saved(self, trans, id=None, revision=None, type=None, config=None, title=None, **kwargs):
        """
        Save (on POST) or load (on GET) a visualization then render.
        """
        # TODO: consider merging saved and render at this point (could break saved URLs, tho)
        if trans.request.method == "POST":
            self._POST_to_saved(trans, id=id, revision=revision, type=type, config=config, title=title, **kwargs)

        # check the id and load the saved visualization
        if id is None:
            return HTTPBadRequest("A valid visualization id is required to load a visualization")
        visualization = self.get_visualization(trans, id, check_ownership=False, check_accessible=True)

        # re-add title to kwargs for passing to render
        if title:
            kwargs["title"] = title
        plugin = self._get_plugin_from_registry(trans, visualization.type)
        try:
            return plugin.render_saved(visualization, trans=trans, **kwargs)
        except Exception as exception:
            self._handle_plugin_error(trans, visualization.type, exception)

    def _POST_to_saved(self, trans, id=None, revision=None, type=None, config=None, title=None, **kwargs):
        """
        Save the visualiztion info (revision, type, config, title, etc.) to
        the Visualization at `id` or to a new Visualization if `id` is None.

        Uses POST/redirect/GET after a successful save, redirecting to GET.
        """
        DEFAULT_VISUALIZATION_NAME = "Unnamed Visualization"

        # post to saved in order to save a visualization
        if type is None or config is None:
            return HTTPBadRequest("A visualization type and config are required to save a visualization")
        if isinstance(config, str):
            config = loads(config)
        title = title or DEFAULT_VISUALIZATION_NAME

        # TODO: allow saving to (updating) a specific revision - should be part of UsesVisualization
        # TODO: would be easier if this returned the visualization directly
        # check security if posting to existing visualization
        if id is not None:
            self.get_visualization(trans, id, check_ownership=True, check_accessible=False)
            # ??: on not owner: error raised, but not returned (status = 200)
        # TODO: there's no security check in save visualization (if passed an id)
        returned = self.save_visualization(trans, config, type, id, title)

        # redirect to GET to prevent annoying 'Do you want to post again?' dialog on page reload
        render_url = web.url_for(controller="visualization", action="saved", id=returned.get("vis_id"))
        return trans.response.send_redirect(render_url)

    #
    # Visualizations.
    #
    @web.expose
    @web.require_login()
    def trackster(self, trans, **kwargs):
        """
        Display browser for the visualization denoted by id and add the datasets listed in `dataset_ids`.
        """

        # define app configuration
        app = {"jscript": "trackster"}

        # get dataset to add
        id = kwargs.get("id", None)

        # get dataset to add
        new_dataset_id = kwargs.get("dataset_id", None)

        # set up new browser if no id provided
        if not id:
            # use dbkey from dataset to be added or from incoming parameter
            dbkey = None
            if new_dataset_id:
                decoded_id = self.decode_id(new_dataset_id)
                hda = self.hda_manager.get_owned(decoded_id, trans.user, current_history=trans.user)
                dbkey = hda.dbkey
                if dbkey == "?":
                    dbkey = kwargs.get("dbkey", None)

            # save database key
            app["default_dbkey"] = dbkey
        else:
            # load saved visualization
            vis = self.get_visualization(trans, id, check_ownership=False, check_accessible=True)
            app["viz_config"] = self.get_visualization_config(trans, vis)

        # backup id
        app["id"] = id

        # add dataset id
        app["add_dataset"] = new_dataset_id

        # check for gene region
        gene_region = GenomeRegion.from_str(kwargs.get("gene_region", ""))

        # update gene region of saved visualization if user parses a new gene region in the url
        if gene_region.chrom is not None:
            app["gene_region"] = {"chrom": gene_region.chrom, "start": gene_region.start, "end": gene_region.end}

        # fill template
        return trans.fill_template("visualization/trackster.mako", config={"app": app, "bundle": "extended"})

    @web.expose
    def circster(self, trans, id=None, hda_ldda=None, dataset_id=None, dbkey=None, **kwargs):
        """
        Display a circster visualization.
        """

        # Get dataset to add.
        dataset = None
        if dataset_id:
            dataset = self.get_hda_or_ldda(trans, hda_ldda, dataset_id)

        # Get/create vis.
        if id:
            # Display existing viz.
            vis = self.get_visualization(trans, id, check_ownership=False, check_accessible=True)
            dbkey = vis.dbkey
        else:
            # Create new viz.
            if not dbkey:
                # If dbkey not specified, use dataset's dbkey.
                dbkey = dataset.dbkey
                if not dbkey or dbkey == "?":
                    # Circster requires a valid dbkey.
                    return trans.show_error_message(
                        "You must set the dataset's dbkey to view it. You can set "
                        "a dataset's dbkey by clicking on the pencil icon and editing "
                        "its attributes.",
                        use_panels=True,
                    )

            vis = self.create_visualization(trans, type="genome", dbkey=dbkey, save=False)

        # Get the vis config and work with it from here on out. Working with the
        # config is only possible because the config structure of trackster/genome
        # visualizations is well known.
        viz_config = self.get_visualization_config(trans, vis)

        # Add dataset if specified.
        if dataset:
            viz_config["tracks"].append(self.get_new_track_config(trans, dataset))

        # Get genome info.
        chroms_info = self.app.genomes.chroms(trans, dbkey=dbkey)
        genome = {"dbkey": dbkey, "chroms_info": chroms_info}

        # Add genome-wide data to each track in viz.
        tracks = viz_config.get("tracks", [])
        for track in tracks:
            dataset_dict = track["dataset"]
            dataset = self.get_hda_or_ldda(trans, dataset_dict["hda_ldda"], dataset_dict["id"])

            genome_data = self._get_genome_data(trans, dataset, dbkey)
            if not isinstance(genome_data, str):
                track["preloaded_data"] = genome_data

        # define app configuration for generic mako template
        app = {"jscript": "circster", "viz_config": viz_config, "genome": genome}

        # fill template
        return trans.fill_template("visualization/trackster.mako", config={"app": app, "bundle": "extended"})

    @web.expose
    def sweepster(self, trans, id=None, hda_ldda=None, dataset_id=None, regions=None, **kwargs):
        """
        Displays a sweepster visualization using the incoming parameters. If id is available,
        get the visualization with the given id; otherwise, create a new visualization using
        a given dataset and regions.
        """
        regions = regions or "{}"
        # Need to create history if necessary in order to create tool form.
        trans.get_history(most_recent=True, create=True)

        if id:
            # Loading a shared visualization.
            viz = self.get_visualization(trans, id)
            viz_config = self.get_visualization_config(trans, viz)
            decoded_id = self.decode_id(viz_config["dataset_id"])
            dataset = self.hda_manager.get_owned(decoded_id, trans.user, current_history=trans.history)
        else:
            # Loading new visualization.
            dataset = self.get_hda_or_ldda(trans, hda_ldda, dataset_id)
            job = self.hda_manager.creating_job(dataset)
            viz_config = {"dataset_id": dataset_id, "tool_id": job.tool_id, "regions": loads(regions)}

        # Add tool, dataset attributes to config based on id.
        tool = trans.app.toolbox.get_tool(viz_config["tool_id"])
        viz_config["tool"] = tool.to_dict(trans, io_details=True)
        viz_config["dataset"] = trans.security.encode_dict_ids(dataset.to_dict())

        return trans.fill_template_mako("visualization/sweepster.mako", config=viz_config)

    def get_item(self, trans, id):
        return self.get_visualization(trans, id)

    @web.expose
    def phyloviz(self, trans, id=None, dataset_id=None, tree_index=0, **kwargs):
        config = None
        data = None

        # if id, then this is a saved visualization; get its config and the dataset_id from there
        if id:
            visualization = self.get_visualization(trans, id)
            config = self.get_visualization_config(trans, visualization)
            dataset_id = config.get("dataset_id", None)

        # get the hda if we can, then its data using the phyloviz parsers
        if dataset_id:
            decoded_id = self.decode_id(dataset_id)
            hda = self.hda_manager.get_accessible(decoded_id, trans.user)
            hda = self.hda_manager.error_if_uploading(hda)
        else:
            return trans.show_message("Phyloviz couldn't find a dataset_id")

        pd = PhylovizDataProvider(original_dataset=hda)
        data = pd.get_data(tree_index=tree_index)

        # ensure at least a default configuration (gen. an new/unsaved visualization)
        if not config:
            config = {
                "dataset_id": dataset_id,
                "title": hda.display_name(),
                "ext": hda.datatype.file_ext,
                "treeIndex": tree_index,
                "saved_visualization": False,
            }
        return trans.fill_template_mako("visualization/phyloviz.mako", data=data, config=config)



import logging
from html.parser import HTMLParser

from markupsafe import escape
from sqlalchemy import desc
from sqlalchemy.orm import (
    joinedload,
    lazyload,
    undefer,
)
from sqlalchemy.sql import expression

from galaxy import (
    model,
    util,
    web,
)
from galaxy.managers.sharable import SlugBuilder
from galaxy.managers.workflows import (
    MissingToolsException,
    WorkflowUpdateOptions,
)
from galaxy.model.base import transaction
from galaxy.model.item_attrs import UsesItemRatings
from galaxy.tools.parameters.basic import workflow_building_modes
from galaxy.util import FILENAME_VALID_CHARS
from galaxy.util.sanitize_html import sanitize_html
from galaxy.web import url_for
from galaxy.web.framework.helpers import (
    grids,
    time_ago,
)
from galaxy.webapps.base.controller import (
    BaseUIController,
    SharableMixin,
    UsesStoredWorkflowMixin,
)
from galaxy.workflow.extract import (
    extract_workflow,
    summarize,
)
from galaxy.workflow.modules import (
    load_module_sections,
    module_factory,
)
from galaxy.workflow.render import (
    STANDALONE_SVG_TEMPLATE,
    WorkflowCanvas,
)

log = logging.getLogger(__name__)


class StoredWorkflowListGrid(grids.Grid):
    class StepsColumn(grids.GridColumn):
        def get_value(self, trans, grid, workflow):
            return len(workflow.latest_workflow.steps)

    # Grid definition
    use_panels = True
    title = "Saved Workflows"
    model_class = model.StoredWorkflow
    default_filter = {"name": "All", "tags": "All"}
    default_sort_key = "-update_time"
    columns = [
        grids.TextColumn("Name", key="name", attach_popup=True, filterable="advanced"),
        grids.IndividualTagsColumn(
            "Tags",
            "tags",
            model_tag_association_class=model.StoredWorkflowTagAssociation,
            filterable="advanced",
            grid_name="StoredWorkflowListGrid",
        ),
        StepsColumn("Steps"),
        grids.GridColumn("Created", key="create_time", format=time_ago),
        grids.GridColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    operations = [
        grids.GridOperation(
            "Edit", allow_multiple=False, condition=(lambda item: not item.deleted), async_compatible=False
        ),
        grids.GridOperation("Run", condition=(lambda item: not item.deleted), async_compatible=False),
        grids.GridOperation("Copy", condition=(lambda item: not item.deleted), async_compatible=False),
        grids.GridOperation("Rename", condition=(lambda item: not item.deleted), async_compatible=False),
        grids.GridOperation("Sharing", condition=(lambda item: not item.deleted), async_compatible=False),
        grids.GridOperation("Delete", condition=(lambda item: item.deleted), async_compatible=True),
    ]

    def apply_query_filter(self, trans, query, **kwargs):
        return query.filter_by(user=trans.user, deleted=False)


class StoredWorkflowAllPublishedGrid(grids.Grid):
    title = "Published Workflows"
    model_class = model.StoredWorkflow
    default_sort_key = "update_time"
    default_filter = dict(public_url="All", username="All", tags="All")
    columns = [
        grids.PublicURLColumn("Name", key="name", filterable="advanced", attach_popup=True),
        grids.OwnerAnnotationColumn(
            "Annotation",
            key="annotation",
            model_annotation_association_class=model.StoredWorkflowAnnotationAssociation,
            filterable="advanced",
        ),
        grids.OwnerColumn("Owner", key="username", model_class=model.User, filterable="advanced"),
        grids.CommunityRatingColumn("Community Rating", key="rating"),
        grids.CommunityTagsColumn(
            "Community Tags",
            key="tags",
            model_tag_association_class=model.StoredWorkflowTagAssociation,
            filterable="advanced",
            grid_name="PublicWorkflowListGrid",
        ),
        grids.ReverseSortColumn("Last Updated", key="update_time", format=time_ago),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search name, annotation, owner, and tags",
            cols_to_filter=[columns[0], columns[1], columns[2], columns[4]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    operations = [
        grids.GridOperation(
            "Run",
            condition=(lambda item: not item.deleted),
            allow_multiple=False,
            url_args=dict(controller="workflows", action="run"),
        ),
        grids.GridOperation(
            "Import", condition=(lambda item: not item.deleted), allow_multiple=False, url_args=dict(action="imp")
        ),
        grids.GridOperation(
            "Save as File",
            condition=(lambda item: not item.deleted),
            allow_multiple=False,
            url_args=dict(action="export_to_file"),
        ),
    ]
    num_rows_per_page = 50
    use_paging = True

    def build_initial_query(self, trans, **kwargs):
        # See optimization description comments and TODO for tags in matching public histories query.
        # In addition to that - be sure to lazyload the latest_workflow - it isn't needed and it causes all
        # of its steps to be eagerly loaded.
        return (
            trans.sa_session.query(self.model_class)
            .join(self.model_class.user)
            .options(
                lazyload(self.model_class.latest_workflow),
                joinedload(self.model_class.user).load_only(model.User.username),
                joinedload(self.model_class.annotations),
                undefer(self.model_class.average_rating),
            )
        )

    def apply_query_filter(self, trans, query, **kwargs):
        # A public workflow is published, has a slug, and is not deleted.
        return (
            query.filter(self.model_class.published == expression.true())
            .filter(self.model_class.slug.isnot(None))
            .filter(self.model_class.deleted == expression.false())
        )


# Simple HTML parser to get all content in a single tag.
class SingleTagContentsParser(HTMLParser):
    def __init__(self, target_tag):
        # Cannot use super() because HTMLParser is an old-style class in Python2
        HTMLParser.__init__(self)
        self.target_tag = target_tag
        self.cur_tag = None
        self.tag_content = ""

    def handle_starttag(self, tag, attrs):
        """Called for each start tag."""
        self.cur_tag = tag

    def handle_data(self, text):
        """Called for each block of plain text."""
        if self.cur_tag == self.target_tag:
            self.tag_content += text


class WorkflowController(BaseUIController, SharableMixin, UsesStoredWorkflowMixin, UsesItemRatings):
    stored_list_grid = StoredWorkflowListGrid()
    published_list_grid = StoredWorkflowAllPublishedGrid()
    slug_builder = SlugBuilder()

    @web.expose
    @web.require_login("use Galaxy workflows")
    def list_grid(self, trans, **kwargs):
        """List user's stored workflows."""
        # status = message = None
        if "operation" in kwargs:
            operation = kwargs["operation"].lower()
            if operation == "rename":
                return self.rename(trans, **kwargs)
            workflow_ids = util.listify(kwargs.get("id", []))
            if operation == "sharing":
                return self.sharing(trans, id=workflow_ids)
        return self.stored_list_grid(trans, **kwargs)

    @web.expose
    @web.json
    def list_published(self, trans, **kwargs):
        return self.published_list_grid(trans, **kwargs)

    @web.expose
    def display_by_username_and_slug(self, trans, username, slug, format="html", **kwargs):
        """
        Display workflow based on a username and slug. Format can be html, json, or json-download.
        """

        # Get workflow by username and slug. Security is handled by the display methods below.
        session = trans.sa_session
        user = session.query(model.User).filter_by(username=username).first()
        if not user:
            raise web.httpexceptions.HTTPNotFound()
        stored_workflow = (
            trans.sa_session.query(model.StoredWorkflow).filter_by(user=user, slug=slug, deleted=False).first()
        )
        if not stored_workflow:
            raise web.httpexceptions.HTTPNotFound()
        encoded_id = trans.security.encode_id(stored_workflow.id)

        # Display workflow in requested format.
        if format == "html":
            return self._display(trans, stored_workflow)
        elif format == "json":
            return self.for_direct_import(trans, encoded_id)
        elif format == "json-download":
            return self.export_to_file(trans, encoded_id)

    def _display(self, trans, stored_workflow):
        """Diplay workflow in client."""
        if stored_workflow is None:
            raise web.httpexceptions.HTTPNotFound()

        # Security check raises error if user cannot access workflow.
        self.security_check(trans, stored_workflow, False, True)

        # Get data for workflow's steps.
        self.get_stored_workflow_steps(trans, stored_workflow)

        # Get annotations.
        stored_workflow.annotation = self.get_item_annotation_str(
            trans.sa_session, stored_workflow.user, stored_workflow
        )
        for step in stored_workflow.latest_workflow.steps:
            step.annotation = self.get_item_annotation_str(trans.sa_session, stored_workflow.user, step)

        # Encode page identifier.
        workflow_id = trans.security.encode_id(stored_workflow.id)

        # Redirect to client.
        return trans.response.send_redirect(
            web.url_for(
                controller="published",
                action="workflow",
                id=workflow_id,
            )
        )

    @web.expose
    @web.require_login("to import a workflow", use_panels=True)
    def imp(self, trans, id, **kwargs):
        """Imports a workflow shared by other users."""
        # Set referer message.
        referer = trans.request.referer
        if referer and not referer.startswith(f"{trans.request.application_url}{url_for('/login')}"):
            referer_message = f"<a href='{escape(referer)}'>return to the previous page</a>"
        else:
            referer_message = f"<a href='{url_for('/')}'>go to Galaxy's start page</a>"

        # Do import.
        stored = self.get_stored_workflow(trans, id, check_ownership=False)
        if stored.importable is False:
            return trans.show_error_message(
                f"The owner of this workflow has disabled imports via this link.<br>You can {referer_message}",
                use_panels=True,
            )
        elif stored.deleted:
            return trans.show_error_message(
                f"You can't import this workflow because it has been deleted.<br>You can {referer_message}",
                use_panels=True,
            )
        self._import_shared_workflow(trans, stored)

        # Redirect to load galaxy frames.
        return trans.show_ok_message(
            message="""Workflow "{}" has been imported. <br>You can <a href="{}">start using this workflow</a> or {}.""".format(
                stored.name, web.url_for("/workflows/list"), referer_message
            )
        )

    @web.expose
    @web.require_login("use Galaxy workflows")
    def rename_async(self, trans, id, new_name=None, **kwargs):
        stored = self.get_stored_workflow(trans, id)
        if new_name:
            san_new_name = sanitize_html(new_name)
            stored.name = san_new_name
            stored.latest_workflow.name = san_new_name
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            return stored.name

    @web.expose
    @web.require_login("use Galaxy workflows")
    def annotate_async(self, trans, id, new_annotation=None, **kwargs):
        stored = self.get_stored_workflow(trans, id)
        if new_annotation:
            # Sanitize annotation before adding it.
            new_annotation = sanitize_html(new_annotation)
            self.add_item_annotation(trans.sa_session, trans.get_user(), stored, new_annotation)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            return new_annotation

    @web.expose
    @web.require_login("use Galaxy workflows")
    def gen_image(self, trans, id, **kwargs):
        stored = self.get_stored_workflow(trans, id, check_ownership=True)
        try:
            svg = self._workflow_to_svg_canvas(trans, stored)
        except Exception:
            message = (
                "Galaxy is unable to create the SVG image. Please check your workflow, there might be missing tools."
            )
            return trans.show_error_message(message)
        trans.response.set_content_type("image/svg+xml")
        s = STANDALONE_SVG_TEMPLATE % svg.tostring()
        return s.encode("utf-8")

    @web.legacy_expose_api
    def create(self, trans, payload=None, **kwd):
        if trans.request.method == "GET":
            return {
                "title": "Create Workflow",
                "inputs": [
                    {"name": "workflow_name", "label": "Name", "value": "Unnamed workflow"},
                    {
                        "name": "workflow_annotation",
                        "label": "Annotation",
                        "help": "A description of the workflow; annotation is shown alongside shared or published workflows.",
                    },
                ],
            }
        else:
            user = trans.get_user()
            workflow_name = payload.get("workflow_name")
            workflow_annotation = payload.get("workflow_annotation")
            if not workflow_name:
                return self.message_exception(trans, "Please provide a workflow name.")
            # Create the new stored workflow
            stored_workflow = model.StoredWorkflow()
            stored_workflow.name = workflow_name
            stored_workflow.user = user
            self.slug_builder.create_item_slug(trans.sa_session, stored_workflow)
            # And the first (empty) workflow revision
            workflow = model.Workflow()
            workflow.name = workflow_name
            workflow.stored_workflow = stored_workflow
            stored_workflow.latest_workflow = workflow
            # Add annotation.
            workflow_annotation = sanitize_html(workflow_annotation)
            self.add_item_annotation(trans.sa_session, trans.get_user(), stored_workflow, workflow_annotation)
            # Persist
            session = trans.sa_session
            session.add(stored_workflow)
            with transaction(session):
                session.commit()
            return {
                "id": trans.security.encode_id(stored_workflow.id),
                "message": f"Workflow {workflow_name} has been created.",
            }

    @web.json
    def save_workflow_as(
        self, trans, workflow_name, workflow_data, workflow_annotation="", from_tool_form=False, **kwargs
    ):
        """
        Creates a new workflow based on Save As command. It is a new workflow, but
        is created with workflow_data already present.
        """
        user = trans.get_user()
        if workflow_name is not None:
            workflow_contents_manager = self.app.workflow_contents_manager
            stored_workflow = model.StoredWorkflow()
            stored_workflow.name = workflow_name
            stored_workflow.user = user
            self.slug_builder.create_item_slug(trans.sa_session, stored_workflow)
            workflow = model.Workflow()
            workflow.name = workflow_name
            workflow.stored_workflow = stored_workflow
            stored_workflow.latest_workflow = workflow
            # Add annotation.
            workflow_annotation = sanitize_html(workflow_annotation)
            self.add_item_annotation(trans.sa_session, trans.get_user(), stored_workflow, workflow_annotation)

            # Persist
            session = trans.sa_session
            session.add(stored_workflow)
            with transaction(session):
                session.commit()
            workflow_update_options = WorkflowUpdateOptions(
                update_stored_workflow_attributes=False,  # taken care of above
                from_tool_form=from_tool_form,
            )
            try:
                workflow, errors = workflow_contents_manager.update_workflow_from_raw_description(
                    trans,
                    stored_workflow,
                    workflow_data,
                    workflow_update_options,
                )
            except MissingToolsException as e:
                return dict(
                    name=e.workflow.name,
                    message=(
                        "This workflow includes missing or invalid tools. "
                        "It cannot be saved until the following steps are removed or the missing tools are enabled."
                    ),
                    errors=e.errors,
                )
            return trans.security.encode_id(stored_workflow.id)
        else:
            # This is an error state, 'save as' must have a workflow_name
            log.exception("Error in Save As workflow: no name.")

    @web.expose
    @web.json
    @web.require_login("edit workflows")
    def editor(self, trans, id=None, workflow_id=None, version=None, **kwargs):
        """
        Render the main workflow editor interface. The canvas is embedded as
        an iframe (necessary for scrolling to work properly), which is
        rendered by `editor_canvas`.
        """
        if not id:
            if workflow_id:
                stored_workflow = self.app.workflow_manager.get_stored_workflow(trans, workflow_id, by_stored_id=False)
                self.security_check(trans, stored_workflow, True, False)
                id = trans.security.encode_id(stored_workflow.id)
        stored = self.get_stored_workflow(trans, id)
        # The following query loads all user-owned workflows,
        # So that they can be copied or inserted in the workflow editor.
        workflows = (
            trans.sa_session.query(model.StoredWorkflow)
            .filter_by(user=trans.user, deleted=False, hidden=False)
            .order_by(desc(model.StoredWorkflow.table.c.update_time))
            .options(joinedload(model.StoredWorkflow.latest_workflow).joinedload(model.Workflow.steps))
            .all()
        )
        if version is None:
            version = len(stored.workflows) - 1
        else:
            version = int(version)

        # create workflow module models
        module_sections = []
        for module_section in load_module_sections(trans).values():
            module_sections.append(
                {
                    "title": module_section.get("title"),
                    "name": module_section.get("name"),
                    "elems": [
                        {"name": elem.get("name"), "title": elem.get("title"), "description": elem.get("description")}
                        for elem in module_section.get("modules")
                    ],
                }
            )

        # create data manager tool models
        data_managers = []
        if trans.user_is_admin and trans.app.data_managers.data_managers:
            for data_manager_val in trans.app.data_managers.data_managers.values():
                tool = data_manager_val.tool
                if not tool.hidden:
                    data_managers.append(
                        {
                            "id": tool.id,
                            "name": tool.name,
                            "hidden": tool.hidden,
                            "description": tool.description,
                            "is_workflow_compatible": tool.is_workflow_compatible,
                        }
                    )

        # create workflow models
        workflows = [
            {
                "id": trans.security.encode_id(workflow.id),
                "latest_id": trans.security.encode_id(workflow.latest_workflow.id),
                "step_count": len(workflow.latest_workflow.steps),
                "name": workflow.name,
            }
            for workflow in workflows
            if workflow.id != stored.id
        ]

        # identify item tags
        item_tags = stored.make_tag_string_list()

        # build workflow editor model
        editor_config = {
            "id": trans.security.encode_id(stored.id),
            "name": stored.name,
            "tags": item_tags,
            "initialVersion": version,
            "annotation": self.get_item_annotation_str(trans.sa_session, trans.user, stored),
            "moduleSections": module_sections,
            "dataManagers": data_managers,
            "workflows": workflows,
        }

        # parse to mako
        return editor_config

    @web.json
    def load_workflow(self, trans, id, version=None, **kwargs):
        """
        Get the latest Workflow for the StoredWorkflow identified by `id` and
        encode it as a json string that can be read by the workflow editor
        web interface.
        """
        trans.workflow_building_mode = workflow_building_modes.ENABLED
        stored = self.get_stored_workflow(trans, id, check_ownership=True, check_accessible=False)
        workflow_contents_manager = self.app.workflow_contents_manager
        return workflow_contents_manager.workflow_to_dict(trans, stored, style="editor", version=version)

    @web.json_pretty
    def for_direct_import(self, trans, id, **kwargs):
        """
        Get the latest Workflow for the StoredWorkflow identified by `id` and
        encode it as a json string that can be imported back into Galaxy

        This has slightly different information than the above. In particular,
        it does not attempt to decode forms and build UIs, it just stores
        the raw state.
        """
        stored = self.get_stored_workflow(trans, id, check_ownership=False, check_accessible=True)
        return self._workflow_to_dict(trans, stored)

    @web.json_pretty
    def export_to_file(self, trans, id):
        """
        Get the latest Workflow for the StoredWorkflow identified by `id` and
        export it to a JSON file that can be imported back into Galaxy.

        This has slightly different information than the above. In particular,
        it does not attempt to decode forms and build UIs, it just stores
        the raw state.
        """

        # Get workflow.
        stored = self.get_stored_workflow(trans, id, check_ownership=False, check_accessible=True)

        # Stream workflow to file.
        stored_dict = self._workflow_to_dict(trans, stored)
        if not stored_dict:
            # This workflow has a tool that's missing from the distribution
            trans.response.status = 400
            return "Workflow cannot be exported due to missing tools."
        sname = stored.name
        sname = "".join(c in FILENAME_VALID_CHARS and c or "_" for c in sname)[0:150]
        trans.response.headers["Content-Disposition"] = f'attachment; filename="Galaxy-Workflow-{sname}.ga"'
        trans.response.set_content_type("application/galaxy-archive")
        return stored_dict

    @web.expose
    def build_from_current_history(
        self,
        trans,
        job_ids=None,
        dataset_ids=None,
        dataset_collection_ids=None,
        workflow_name=None,
        dataset_names=None,
        dataset_collection_names=None,
        **kwargs,
    ):
        user = trans.get_user()
        history = trans.get_history()
        if not user:
            return trans.show_error_message("Must be logged in to create workflows")
        if (job_ids is None and dataset_ids is None) or workflow_name is None:
            jobs, warnings = summarize(trans)
            # Render
            return trans.fill_template(
                "workflow/build_from_current_history.mako", jobs=jobs, warnings=warnings, history=history
            )
        else:
            # If there is just one dataset name selected or one dataset collection, these
            # come through as string types instead of lists. xref #3247.
            dataset_names = util.listify(dataset_names)
            dataset_collection_names = util.listify(dataset_collection_names)
            stored_workflow = extract_workflow(
                trans,
                user=user,
                job_ids=job_ids,
                dataset_ids=dataset_ids,
                dataset_collection_ids=dataset_collection_ids,
                workflow_name=workflow_name,
                dataset_names=dataset_names,
                dataset_collection_names=dataset_collection_names,
            )
            # Index page with message
            workflow_id = trans.security.encode_id(stored_workflow.id)
            edit_url = url_for(f"/workflows/edit?id={workflow_id}")
            run_url = url_for(f"/workflows/run?id={workflow_id}")
            return trans.show_message(
                f'Workflow "{escape(workflow_name)}" created from current history. '
                f'You can <a href="{edit_url}" target="_parent">edit</a> or <a href="{run_url}" target="_parent">run</a> the workflow.'
            )

    def get_item(self, trans, id):
        return self.get_stored_workflow(trans, id)

    def _workflow_to_svg_canvas(self, trans, stored):
        workflow = stored.latest_workflow
        workflow_canvas = WorkflowCanvas()
        for step in workflow.steps:
            # Load from database representation
            module = module_factory.from_workflow_step(trans, step)
            module_name = module.get_name()
            module_data_inputs = module.get_data_inputs()
            module_data_outputs = module.get_data_outputs()
            workflow_canvas.populate_data_for_step(
                step,
                module_name,
                module_data_inputs,
                module_data_outputs,
            )
        workflow_canvas.add_steps()
        return workflow_canvas.finish()



"""
Manager and Serializer for TS groups.
"""
import logging

from sqlalchemy import (
    false,
    true,
)
from sqlalchemy.orm.exc import (
    MultipleResultsFound,
    NoResultFound,
)

from galaxy.exceptions import (
    Conflict,
    InconsistentDatabase,
    InternalServerError,
    ItemAccessibilityException,
    ObjectNotFound,
    RequestParameterInvalidException,
)
from galaxy.model.base import transaction

log = logging.getLogger(__name__)


# =============================================================================
class GroupManager:
    """
    Interface/service object for interacting with TS groups.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, trans, decoded_group_id=None, name=None):
        """
        Get the group from the DB based on its ID or name.

        :param  decoded_group_id:       decoded group id
        :type   decoded_group_id:       int

        :returns:   the requested group
        :rtype:     tool_shed.model.Group
        """
        if decoded_group_id is None and name is None:
            raise RequestParameterInvalidException("You must supply either ID or a name of the group.")

        name_query = trans.sa_session.query(trans.app.model.Group).filter(trans.app.model.Group.table.c.name == name)
        id_query = trans.sa_session.query(trans.app.model.Group).filter(
            trans.app.model.Group.table.c.id == decoded_group_id
        )

        try:
            group = id_query.one() if decoded_group_id else name_query.one()
        except MultipleResultsFound:
            raise InconsistentDatabase("Multiple groups found with the same identifier.")
        except NoResultFound:
            raise ObjectNotFound("No group found with the identifier provided.")
        except Exception:
            raise InternalServerError("Error loading from the database.")
        return group

    def create(self, trans, name, description=""):
        """
        Create a new group.
        """
        if not trans.user_is_admin:
            raise ItemAccessibilityException("Only administrators can create groups.")
        else:
            if self.get(trans, name=name):
                raise Conflict(f"Group with the given name already exists. Name: {str(name)}")
            # TODO add description field to the model
            group = trans.app.model.Group(name=name)
            trans.sa_session.add(group)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
            return group

    def update(self, trans, group, name=None, description=None):
        """
        Update the given group
        """
        changed = False
        if not trans.user_is_admin:
            raise ItemAccessibilityException("Only administrators can update groups.")
        if group.deleted:
            raise RequestParameterInvalidException("You cannot modify a deleted group. Undelete it first.")
        if name is not None:
            group.name = name
            changed = True
        if description is not None:
            group.description = description
            changed = True
        if changed:
            trans.sa_session.add(group)
            with transaction(trans.sa_session):
                trans.sa_session.commit()
        return group

    def delete(self, trans, group, undelete=False):
        """
        Mark given group deleted/undeleted based on the flag.
        """
        if not trans.user_is_admin:
            raise ItemAccessibilityException("Only administrators can delete and undelete groups.")
        if undelete:
            group.deleted = False
        else:
            group.deleted = True
        trans.sa_session.add(group)
        with transaction(trans.sa_session):
            trans.sa_session.commit()
        return group

    def list(self, trans, deleted=False):
        """
        Return a list of groups from the DB.

        :returns: query that will emit all groups
        :rtype:   sqlalchemy query
        """
        is_admin = trans.user_is_admin
        query = trans.sa_session.query(trans.app.model.Group)
        if is_admin:
            if deleted is None:
                #  Flag is not specified, do not filter on it.
                pass
            elif deleted:
                query = query.filter(trans.app.model.Group.table.c.deleted == true())
            else:
                query = query.filter(trans.app.model.Group.table.c.deleted == false())
        else:
            query = query.filter(trans.app.model.Group.table.c.deleted == false())
        return query



import logging
import os
import shutil
import tarfile
import tempfile

import requests

from galaxy import (
    util,
    web,
)
from galaxy.tool_shed.util import dependency_display
from galaxy.util import checkers
from galaxy.webapps.base.controller import BaseUIController
from tool_shed.dependencies import attribute_handlers
from tool_shed.metadata import repository_metadata_manager
from tool_shed.repository_types import util as rt_util
from tool_shed.tools.data_table_manager import ShedToolDataTableManager
from tool_shed.util import (
    basic_util,
    commit_util,
    hg_util,
    repository_content_util,
    repository_util,
    shed_util_common as suc,
    xml_util,
)
from tool_shed.util.web_util import escape
from tool_shed.webapp.framework.decorators import require_login

log = logging.getLogger(__name__)


class UploadController(BaseUIController):
    @web.expose
    @require_login("upload", use_panels=True)
    def upload(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        commit_message = escape(kwd.get("commit_message", "Uploaded"))
        repository_id = kwd.get("repository_id", "")
        repository = repository_util.get_repository_in_tool_shed(trans.app, repository_id)
        repo_dir = repository.repo_path(trans.app)
        uncompress_file = util.string_as_bool(kwd.get("uncompress_file", "true"))
        remove_repo_files_not_in_tar = util.string_as_bool(kwd.get("remove_repo_files_not_in_tar", "true"))
        uploaded_file = None
        upload_point = commit_util.get_upload_point(repository, **kwd)
        tip = repository.tip()
        file_data = kwd.get("file_data", "")
        url = kwd.get("url", "")
        # Part of the upload process is sending email notification to those that have registered to
        # receive them.  One scenario occurs when the first change set is produced for the repository.
        # See the suc.handle_email_alerts() method for the definition of the scenarios.
        new_repo_alert = repository.is_new()
        uploaded_directory = None
        if kwd.get("upload_button", False):
            if file_data == "" and url == "":
                message = "No files were entered on the upload form."
                status = "error"
                uploaded_file = None
            elif url and url.startswith("hg"):
                # Use mercurial clone to fetch repository, contents will then be copied over.
                uploaded_directory = tempfile.mkdtemp()
                repo_url = f"http{url[len('hg'):]}"
                cloned_ok, error_message = hg_util.clone_repository(repo_url, uploaded_directory)
                if not cloned_ok:
                    message = f"Error uploading via mercurial clone: {error_message}"
                    status = "error"
                    basic_util.remove_dir(uploaded_directory)
                    uploaded_directory = None
            elif url:
                valid_url = True
                try:
                    stream = requests.get(url, stream=True, timeout=util.DEFAULT_SOCKET_TIMEOUT)
                except Exception as e:
                    valid_url = False
                    message = f"Error uploading file via http: {util.unicodify(e)}"
                    status = "error"
                    uploaded_file = None
                if valid_url:
                    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as uploaded_file:
                        uploaded_file_name = uploaded_file.name
                        for chunk in stream.iter_content(chunk_size=util.CHUNK_SIZE):
                            if chunk:
                                uploaded_file.write(chunk)
                        uploaded_file.flush()
                    uploaded_file_filename = url.split("/")[-1]
                    isempty = os.path.getsize(os.path.abspath(uploaded_file_name)) == 0
            elif file_data not in ("", None):
                uploaded_file = file_data.file
                uploaded_file_name = uploaded_file.name
                uploaded_file_filename = os.path.split(file_data.filename)[-1]
                isempty = os.path.getsize(os.path.abspath(uploaded_file_name)) == 0
            if uploaded_file or uploaded_directory:
                rdah = attribute_handlers.RepositoryDependencyAttributeHandler(trans.app, unpopulate=False)
                tdah = attribute_handlers.ToolDependencyAttributeHandler(trans.app, unpopulate=False)
                stdtm = ShedToolDataTableManager(trans.app)
                ok = True
                isgzip = False
                isbz2 = False
                if uploaded_file:
                    if uncompress_file:
                        isgzip = checkers.is_gzip(uploaded_file_name)
                        if not isgzip:
                            isbz2 = checkers.is_bz2(uploaded_file_name)
                    if isempty:
                        tar = None
                        istar = False
                    else:
                        # Determine what we have - a single file or an archive
                        try:
                            if (isgzip or isbz2) and uncompress_file:
                                # Open for reading with transparent compression.
                                tar = tarfile.open(uploaded_file_name, "r:*")
                            else:
                                tar = tarfile.open(uploaded_file_name)
                            istar = True
                        except tarfile.ReadError:
                            tar = None
                            istar = False
                else:
                    # Uploaded directory
                    istar = False
                if istar:
                    (
                        ok,
                        message,
                        files_to_remove,
                        content_alert_str,
                        undesirable_dirs_removed,
                        undesirable_files_removed,
                    ) = repository_content_util.upload_tar(
                        trans,
                        rdah,
                        tdah,
                        repository,
                        tar,
                        uploaded_file,
                        upload_point,
                        remove_repo_files_not_in_tar,
                        commit_message,
                        new_repo_alert,
                    )
                elif uploaded_directory:
                    (
                        ok,
                        message,
                        files_to_remove,
                        content_alert_str,
                        undesirable_dirs_removed,
                        undesirable_files_removed,
                    ) = self.upload_directory(
                        trans,
                        rdah,
                        tdah,
                        repository,
                        uploaded_directory,
                        upload_point,
                        remove_repo_files_not_in_tar,
                        commit_message,
                        new_repo_alert,
                    )
                else:
                    if (isgzip or isbz2) and uncompress_file:
                        uploaded_file_filename = commit_util.uncompress(
                            repository, uploaded_file_name, uploaded_file_filename, isgzip=isgzip, isbz2=isbz2
                        )
                    if (
                        repository.type == rt_util.REPOSITORY_SUITE_DEFINITION
                        and uploaded_file_filename != rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME
                    ):
                        ok = False
                        message = "Repositories of type <b>Repository suite definition</b> can only contain a single file named "
                        message += "<b>repository_dependencies.xml</b>."
                    elif (
                        repository.type == rt_util.TOOL_DEPENDENCY_DEFINITION
                        and uploaded_file_filename != rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME
                    ):
                        ok = False
                        message = "Repositories of type <b>Tool dependency definition</b> can only contain a single file named "
                        message += "<b>tool_dependencies.xml</b>."
                    if ok:
                        if upload_point is not None:
                            full_path = os.path.abspath(os.path.join(repo_dir, upload_point, uploaded_file_filename))
                        else:
                            full_path = os.path.abspath(os.path.join(repo_dir, uploaded_file_filename))
                        # Move some version of the uploaded file to the load_point within the repository hierarchy.
                        if uploaded_file_filename in [rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME]:
                            # Inspect the contents of the file to see if toolshed or changeset_revision attributes
                            # are missing and if so, set them appropriately.
                            altered, root_elem, error_message = rdah.handle_tag_attributes(uploaded_file_name)
                            if error_message:
                                ok = False
                                message = error_message
                                status = "error"
                            elif altered:
                                tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                                shutil.move(tmp_filename, full_path)
                            else:
                                shutil.move(uploaded_file_name, full_path)
                        elif uploaded_file_filename in [rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME]:
                            # Inspect the contents of the file to see if changeset_revision values are
                            # missing and if so, set them appropriately.
                            altered, root_elem, error_message = tdah.handle_tag_attributes(uploaded_file_name)
                            if error_message:
                                ok = False
                                message = error_message
                                status = "error"
                            if ok:
                                if altered:
                                    tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                                    shutil.move(tmp_filename, full_path)
                                else:
                                    shutil.move(uploaded_file_name, full_path)
                        else:
                            shutil.move(uploaded_file_name, full_path)
                        if ok:
                            # See if any admin users have chosen to receive email alerts when a repository is updated.
                            # If so, check every uploaded file to ensure content is appropriate.
                            check_contents = commit_util.check_file_contents_for_email_alerts(trans.app)
                            if check_contents and os.path.isfile(full_path):
                                content_alert_str = commit_util.check_file_content_for_html_and_images(full_path)
                            else:
                                content_alert_str = ""
                            hg_util.add_changeset(repo_dir, full_path)
                            hg_util.commit_changeset(
                                repo_dir,
                                full_path_to_changeset=full_path,
                                username=trans.user.username,
                                message=commit_message,
                            )
                            if full_path.endswith("tool_data_table_conf.xml.sample"):
                                # Handle the special case where a tool_data_table_conf.xml.sample file is being uploaded
                                # by parsing the file and adding new entries to the in-memory trans.app.tool_data_tables
                                # dictionary.
                                error, error_message = stdtm.handle_sample_tool_data_table_conf_file(
                                    full_path, persist=False
                                )
                                if error:
                                    message = f"{message}<br/>{error_message}"
                            # See if the content of the change set was valid.
                            admin_only = len(repository.downloadable_revisions) != 1
                            suc.handle_email_alerts(
                                trans.app,
                                trans.request.host,
                                repository,
                                content_alert_str=content_alert_str,
                                new_repo_alert=new_repo_alert,
                                admin_only=admin_only,
                            )
                if ok:
                    # Update the repository files for browsing.
                    hg_util.update_repository(repo_dir)
                    # Get the new repository tip.
                    if tip == repository.tip():
                        message = "No changes to repository.  "
                        status = "warning"
                    else:
                        if (isgzip or isbz2) and uncompress_file:
                            uncompress_str = " uncompressed and "
                        else:
                            uncompress_str = " "
                        if uploaded_directory:
                            source_type = "repository"
                            source = url
                        else:
                            source_type = "file"
                            source = uploaded_file_filename
                        message = f"The {source_type} <b>{escape(source)}</b> has been successfully{uncompress_str}uploaded to the repository.  "
                        if istar and (undesirable_dirs_removed or undesirable_files_removed):
                            items_removed = undesirable_dirs_removed + undesirable_files_removed
                            message += (
                                "  %d undesirable items (.hg .svn .git directories, .DS_Store, hgrc files, etc) "
                                % items_removed
                            )
                            message += "were removed from the archive.  "
                        if istar and remove_repo_files_not_in_tar and files_to_remove:
                            if upload_point is not None:
                                message += (
                                    "  %d files were removed from the repository relative to the selected upload point '%s'.  "
                                    % (len(files_to_remove), upload_point)
                                )
                            else:
                                message += "  %d files were removed from the repository root.  " % len(files_to_remove)
                        rmm = repository_metadata_manager.RepositoryMetadataManager(
                            app=trans.app, user=trans.user, repository=repository
                        )
                        status, error_message = rmm.set_repository_metadata_due_to_new_tip(
                            trans.request.host, content_alert_str=content_alert_str, **kwd
                        )
                        if error_message:
                            message = error_message
                        kwd["message"] = message
                    if repository.metadata_revisions:
                        # A repository's metadata revisions are order descending by update_time, so the zeroth revision
                        # will be the tip just after an upload.
                        metadata_dict = repository.metadata_revisions[0].metadata
                    else:
                        metadata_dict = {}
                    dd = dependency_display.DependencyDisplayer(trans.app)
                    if str(repository.type) not in [
                        rt_util.REPOSITORY_SUITE_DEFINITION,
                        rt_util.TOOL_DEPENDENCY_DEFINITION,
                    ]:
                        change_repository_type_message = rt_util.generate_message_for_repository_type_change(
                            trans.app, repository
                        )
                        if change_repository_type_message:
                            message += change_repository_type_message
                            status = "warning"
                        else:
                            # Provide a warning message if a tool_dependencies.xml file is provided, but tool dependencies
                            # weren't loaded due to a requirement tag mismatch or some other problem.  Tool dependency
                            # definitions can define orphan tool dependencies (no relationship to any tools contained in the
                            # repository), so warning messages are important because orphans are always valid.  The repository
                            # owner must be warned in case they did not intend to define an orphan dependency, but simply
                            # provided incorrect information (tool shed, name owner, changeset_revision) for the definition.
                            orphan_message = dd.generate_message_for_orphan_tool_dependencies(repository, metadata_dict)
                            if orphan_message:
                                message += orphan_message
                                status = "warning"
                    # Handle messaging for invalid tool dependencies.
                    invalid_tool_dependencies_message = dd.generate_message_for_invalid_tool_dependencies(metadata_dict)
                    if invalid_tool_dependencies_message:
                        message += invalid_tool_dependencies_message
                        status = "error"
                    # Handle messaging for invalid repository dependencies.
                    invalid_repository_dependencies_message = dd.generate_message_for_invalid_repository_dependencies(
                        metadata_dict, error_from_tuple=True
                    )
                    if invalid_repository_dependencies_message:
                        message += invalid_repository_dependencies_message
                        status = "error"
                    # Reset the tool_data_tables by loading the empty tool_data_table_conf.xml file.
                    stdtm.reset_tool_data_tables()
                    if uploaded_directory:
                        basic_util.remove_dir(uploaded_directory)
                    trans.response.send_redirect(
                        web.url_for(
                            controller="repository",
                            action="browse_repository",
                            id=repository_id,
                            commit_message="Deleted selected files",
                            message=message,
                            status=status,
                        )
                    )
                else:
                    if uploaded_directory:
                        basic_util.remove_dir(uploaded_directory)
                    status = "error"
                # Reset the tool_data_tables by loading the empty tool_data_table_conf.xml file.
                stdtm.reset_tool_data_tables()
        return trans.fill_template(
            "/webapps/tool_shed/repository/upload.mako",
            repository=repository,
            changeset_revision=tip,
            url=url,
            commit_message=commit_message,
            uncompress_file=uncompress_file,
            remove_repo_files_not_in_tar=remove_repo_files_not_in_tar,
            message=message,
            status=status,
        )

    def upload_directory(
        self,
        trans,
        rdah,
        tdah,
        repository,
        uploaded_directory,
        upload_point,
        remove_repo_files_not_in_tar,
        commit_message,
        new_repo_alert,
    ):
        repo_dir = repository.repo_path(trans.app)
        undesirable_dirs_removed = 0
        undesirable_files_removed = 0
        if upload_point is not None:
            full_path = os.path.abspath(os.path.join(repo_dir, upload_point))
        else:
            full_path = os.path.abspath(repo_dir)
        filenames_in_archive = []
        for root, _dirs, files in os.walk(uploaded_directory):
            for uploaded_file in files:
                relative_path = os.path.normpath(os.path.join(os.path.relpath(root, uploaded_directory), uploaded_file))
                if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                    ok = os.path.basename(uploaded_file) == rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME
                elif repository.type == rt_util.TOOL_DEPENDENCY_DEFINITION:
                    ok = os.path.basename(uploaded_file) == rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME
                else:
                    ok = os.path.basename(uploaded_file) not in commit_util.UNDESIRABLE_FILES
                if ok:
                    for file_path_item in relative_path.split("/"):
                        if file_path_item in commit_util.UNDESIRABLE_DIRS:
                            undesirable_dirs_removed += 1
                            ok = False
                            break
                else:
                    undesirable_files_removed += 1
                if ok:
                    uploaded_file_name = os.path.abspath(os.path.join(root, uploaded_file))
                    if os.path.split(uploaded_file_name)[-1] == rt_util.REPOSITORY_DEPENDENCY_DEFINITION_FILENAME:
                        # Inspect the contents of the file to see if toolshed or changeset_revision
                        # attributes are missing and if so, set them appropriately.
                        altered, root_elem, error_message = rdah.handle_tag_attributes(uploaded_file_name)
                        if error_message:
                            return False, error_message, [], "", [], []
                        elif altered:
                            tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                            shutil.move(tmp_filename, uploaded_file_name)
                    elif os.path.split(uploaded_file_name)[-1] == rt_util.TOOL_DEPENDENCY_DEFINITION_FILENAME:
                        # Inspect the contents of the file to see if toolshed or changeset_revision
                        # attributes are missing and if so, set them appropriately.
                        altered, root_elem, error_message = tdah.handle_tag_attributes(uploaded_file_name)
                        if error_message:
                            return False, error_message, [], "", [], []
                        if altered:
                            tmp_filename = xml_util.create_and_write_tmp_file(root_elem)
                            shutil.move(tmp_filename, uploaded_file_name)
                    repo_path = os.path.join(full_path, relative_path)
                    repo_basedir = os.path.normpath(os.path.join(repo_path, os.path.pardir))
                    if not os.path.exists(repo_basedir):
                        os.makedirs(repo_basedir)
                    if os.path.exists(repo_path):
                        if os.path.isdir(repo_path):
                            shutil.rmtree(repo_path)
                        else:
                            os.remove(repo_path)
                    shutil.move(os.path.join(uploaded_directory, relative_path), repo_path)
                    filenames_in_archive.append(relative_path)
        return commit_util.handle_directory_changes(
            trans.app,
            trans.request.host,
            trans.user.username,
            repository,
            full_path,
            filenames_in_archive,
            remove_repo_files_not_in_tar,
            new_repo_alert,
            commit_message,
            undesirable_dirs_removed,
            undesirable_files_removed,
        )



from enum import Enum
from typing import (
    Any,
    Dict,
)

from galaxy import (
    exceptions,
    model,
)
from galaxy.managers import hdas
from galaxy.managers.context import ProvidesUserContext
from galaxy.managers.jobs import (
    JobManager,
    JobSearch,
    view_show_job,
)
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.schema import JobIndexQueryPayload


class JobIndexViewEnum(str, Enum):
    collection = "collection"
    admin_job_list = "admin_job_list"


class JobIndexPayload(JobIndexQueryPayload):
    view: JobIndexViewEnum = JobIndexViewEnum.collection


class JobsService:
    job_manager: JobManager
    job_search: JobSearch
    hda_manager: hdas.HDAManager

    def __init__(
        self,
        job_manager: JobManager,
        job_search: JobSearch,
        hda_manager: hdas.HDAManager,
    ):
        self.job_manager = job_manager
        self.job_search = job_search
        self.hda_manager = hda_manager

    def show(
        self,
        trans: ProvidesUserContext,
        id: DecodedDatabaseIdField,
        full: bool = False,
    ) -> Dict[str, Any]:
        job = self.job_manager.get_accessible_job(trans, id)
        return view_show_job(trans, job, bool(full))

    def index(
        self,
        trans: ProvidesUserContext,
        payload: JobIndexPayload,
    ):
        security = trans.security
        is_admin = trans.user_is_admin
        if payload.view == JobIndexViewEnum.admin_job_list:
            payload.user_details = True
        user_details = payload.user_details
        if payload.view == JobIndexViewEnum.admin_job_list and not is_admin:
            raise exceptions.AdminRequiredException("Only admins can use the admin_job_list view")
        query = self.job_manager.index_query(trans, payload)
        out = []
        view = payload.view
        for job in query.yield_per(model.YIELD_PER_ROWS):
            job_dict = job.to_dict(view, system_details=is_admin)
            j = security.encode_all_ids(job_dict, True)
            if view == JobIndexViewEnum.admin_job_list:
                j["decoded_job_id"] = job.id
            if user_details:
                j["user_email"] = job.get_user_email()
            out.append(j)

        return out



from logging import getLogger
from typing import Optional

from fastapi import (
    Body,
    Path,
    Query,
)

from galaxy.managers.context import ProvidesHistoryContext
from galaxy.schema.fields import DecodedDatabaseIdField
from galaxy.schema.schema import (
    AnyHDCA,
    CreateNewCollectionPayload,
    DatasetCollectionInstanceType,
    DCESummary,
    HDCADetailed,
)
from galaxy.webapps.galaxy.api import (
    depends,
    DependsOnTrans,
    Router,
)
from galaxy.webapps.galaxy.services.dataset_collections import (
    DatasetCollectionAttributesResult,
    DatasetCollectionContentElements,
    DatasetCollectionsService,
    SuitableConverters,
    UpdateCollectionAttributePayload,
)

log = getLogger(__name__)

router = Router(tags=["dataset collections"])

DatasetCollectionIdPathParam: DecodedDatabaseIdField = Path(
    ..., description="The encoded identifier of the dataset collection."
)


DatasetCollectionElementIdPathParam: DecodedDatabaseIdField = Path(
    ..., description="The encoded identifier of the dataset collection element."
)

InstanceTypeQueryParam: DatasetCollectionInstanceType = Query(
    default="history",
    description="The type of collection instance. Either `history` (default) or `library`.",
)


@router.cbv
class FastAPIDatasetCollections:
    service: DatasetCollectionsService = depends(DatasetCollectionsService)

    @router.post(
        "/api/dataset_collections",
        summary="Create a new dataset collection instance.",
    )
    def create(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        payload: CreateNewCollectionPayload = Body(...),
    ) -> HDCADetailed:
        return self.service.create(trans, payload)

    @router.post(
        "/api/dataset_collections/{id}/copy",
        summary="Copy the given collection datasets to a new collection using a new `dbkey` attribute.",
    )
    def copy(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        id: DecodedDatabaseIdField = Path(..., description="The ID of the dataset collection to copy."),
        payload: UpdateCollectionAttributePayload = Body(...),
    ):
        self.service.copy(trans, id, payload)

    @router.get(
        "/api/dataset_collections/{id}/attributes",
        summary="Returns `dbkey`/`extension` attributes for all the collection elements.",
    )
    def attributes(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        id: DecodedDatabaseIdField = DatasetCollectionIdPathParam,
        instance_type: DatasetCollectionInstanceType = InstanceTypeQueryParam,
    ) -> DatasetCollectionAttributesResult:
        return self.service.attributes(trans, id, instance_type)

    @router.get(
        "/api/dataset_collections/{id}/suitable_converters",
        summary="Returns a list of applicable converters for all datatypes in the given collection.",
    )
    def suitable_converters(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        id: DecodedDatabaseIdField = DatasetCollectionIdPathParam,
        instance_type: DatasetCollectionInstanceType = InstanceTypeQueryParam,
    ) -> SuitableConverters:
        return self.service.suitable_converters(trans, id, instance_type)

    @router.get(
        "/api/dataset_collections/{id}",
        summary="Returns detailed information about the given collection.",
    )
    def show(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        id: DecodedDatabaseIdField = DatasetCollectionIdPathParam,
        instance_type: DatasetCollectionInstanceType = InstanceTypeQueryParam,
    ) -> AnyHDCA:
        return self.service.show(trans, id, instance_type)

    @router.get(
        "/api/dataset_collections/{hdca_id}/contents/{parent_id}",
        name="contents_dataset_collection",
        summary="Returns direct child contents of indicated dataset collection parent ID.",
    )
    def contents(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        hdca_id: DecodedDatabaseIdField = DatasetCollectionIdPathParam,
        parent_id: DecodedDatabaseIdField = Path(
            ...,
            description="Parent collection ID describing what collection the contents belongs to.",
        ),
        instance_type: DatasetCollectionInstanceType = InstanceTypeQueryParam,
        limit: Optional[int] = Query(
            default=None,
            description="The maximum number of content elements to return.",
        ),
        offset: Optional[int] = Query(
            default=None,
            description="The number of content elements that will be skipped before returning.",
        ),
    ) -> DatasetCollectionContentElements:
        return self.service.contents(trans, hdca_id, parent_id, instance_type, limit, offset)

    @router.get("/api/dataset_collection_element/{dce_id}")
    def content(
        self,
        trans: ProvidesHistoryContext = DependsOnTrans,
        dce_id: DecodedDatabaseIdField = DatasetCollectionElementIdPathParam,
    ) -> DCESummary:
        return self.service.dce_content(trans, dce_id)
